# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from unittest.mock import AsyncMock, MagicMock, patch

from . import (
    build_validate_http_signature,
    generate_rsa_public_private_key,
    private_key_to_did_key,
    validate_moo_auth_signature,
)
from .test import public_key
from .types import CryptographicIdentifier


def test_generate_rsa_public_private_key():
    public_key, private_key = generate_rsa_public_private_key()

    assert public_key.startswith("-----BEGIN PUBLIC KEY-----")
    assert public_key.endswith("-----END PUBLIC KEY-----\n")

    assert private_key.startswith("-----BEGIN PRIVATE KEY-----")
    assert private_key.endswith("-----END PRIVATE KEY-----\n")


def test_private_key_to_did_key():
    private_key = "z3u2Yxcowsarethebestcowsarethebestcowsarethebest"

    did_key = private_key_to_did_key(private_key)

    assert did_key == "did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"


@patch("bovine.utils.check_max_offset_now")
async def test_validate_moo_auth_signature_get(mock_offset):
    # https://blog.mymath.rocks/2023-03-15/BIN1_Moo_Authentication_and_Authoriation#appendix-test-data

    mock_offset.return_value = True

    request = MagicMock(
        method="get",
        path="/path/to/resource",
        headers={
            "date": "Wed, 15 Mar 2023 17:28:15 GMT",
            "host": "myhost.tld",
            "authorization": "Moo-Auth-1 did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5",  # noqa F501
            "x-moo-signature": "z5ahdHCbP9aJEsDtvG1MEZpxPzuvGKYcdXdKvMq5YL21Z2umxjs1SopCY2Ap8vZxVjTEf6dYbGuB7mtgcgUyNdBLe",  # noqa F501
        },
    )

    result, _ = await validate_moo_auth_signature(request, "myhost.tld")

    assert result == "did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"


@patch("bovine.utils.check_max_offset_now")
async def test_validate_moo_auth_signature_post(mock_offset):
    # https://blog.mymath.rocks/2023-03-15/BIN1_Moo_Authentication_and_Authoriation#appendix-test-data

    mock_offset.return_value = True

    request = MagicMock(
        method="post",
        path="/path/to/resource",
        headers={
            "date": "Wed, 15 Mar 2023 17:28:15 GMT",
            "host": "myhost.tld",
            "authorization": "Moo-Auth-1 did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5",  # noqa F501
            "digest": "sha-256=MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=",
            "x-moo-signature": "z4vPkJaoaSVQp5DrMb8EvCajJcerW36rsyWDELTWQ3cYmaonnGfb8WHiwH54BShidCcmpoyHjanVRYNrXXXka4jAn",  # noqa F501
        },
        get_data=AsyncMock(return_value='{"cows": "good"}'),
    )

    result, _ = await validate_moo_auth_signature(request, "myhost.tld")

    assert result == "did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"


@patch("bovine.utils.check_max_offset_now")
async def test_build_validate_http_signature(mock_offset):
    async def key_retriever(url):
        return CryptographicIdentifier.from_pem(public_key, "owner")

    mock_offset.return_value = True

    validate_http_signature = build_validate_http_signature(key_retriever)

    request = MagicMock(
        method="get",
        path="/path/to/resource",
        url="https://myhost.tld/path/to/resource",
        headers={
            "date": "Wed, 15 Mar 2023 17:28:15 GMT",
            "host": "myhost.tld",
            "signature": 'keyId="key_id",algorithm="rsa-sha256",headers="(request-target) host date",signature="bwgA3UajpDzM07wM+MUkpGHS/mAhhTu+WDiBL1H28J762jUUuggzP8We5+I0WHypSBCWt8Lap2AbYqG6PfTsdFEbk8VWewX34n7/LUSM/WJeA8AFCEnHYPd8rRib4d0+kQV+U4Ai997XUxs2xi2AruNdkB6SNMkSbnQLoBjYT4tzGaf2I1xZuqjD1HvvW0xa1o/nPt4G5kj0tVmgDrrgeKL2hUJ21UdHcXDNFw9QJWf/O6KyEmbumvppKGFUfdpKSh3n2/fP7O4BxqadCbMYj0oDDOF1H3Pa6GV9EfIeSXXLS9TfyPpkXwbWrNpzUBZQ2hRGzX2jS/JEhpSNSVqv6A=="',  # noqa F501
        },
    )

    assert await validate_http_signature(request)


@patch("bovine.utils.check_max_offset_now")
async def test_build_validate_http_signature_post(mock_offset):
    async def key_retriever(url):
        return CryptographicIdentifier.from_pem(public_key, "owner")

    mock_offset.return_value = True

    validate_http_signature = build_validate_http_signature(key_retriever)

    request = MagicMock(
        method="post",
        path="/path/to/resource",
        url="https://myhost.tld/path/to/resource",
        headers={
            "date": "Wed, 15 Mar 2023 17:28:15 GMT",
            "host": "myhost.tld",
            "digest": "sha-256=MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=",
            "signature": 'keyId="key_id",algorithm="rsa-sha256",headers="(request-target) host date digest",signature="QpeLMEh/Z009DgYcSOkrsKr9zW7Wu7UWTQ398qvueCzaEL/Hxrv9C42U+WY/O34x/385lT+z1I3Bk6qMZTDAZSmKGYB5lZJVBdC20a4D++HzIyE62d6CvqUsQwIv4od/hOmmOQ3HKI69MUOsNrLyjIQxUaLSBz1m/wkLv8iMrh9QwULlqTO8WHVjSJlgVwllRtOOcm3wzo/cr2XyFj4gB8CUd3aG3dqzCfOOoNQvUhMDlxP7UqBnGoF6pnYnTnJUSSHdceWbE09TqXLmAHzauYGSdaHRS3XR/Sje7ET9UaOgHHcPJHIN+4jrBpDp2GM69joXWkZwhE0GOl9q3KBSKA=="',  # noqa F501
        },
        get_data=AsyncMock(return_value=b'{"cows": "good"}'),
    )

    assert await validate_http_signature(request)


@patch("bovine.utils.check_max_offset_now")
async def test_build_validate_http_signature_post_content_digest(mock_offset):
    async def key_retriever(url):
        return CryptographicIdentifier.from_pem(public_key, "owner")

    mock_offset.return_value = True

    validate_http_signature = build_validate_http_signature(key_retriever)

    request = MagicMock(
        method="post",
        path="/test_path",
        url="https://test_domain/test_path",
        headers={
            "date": "Mon, 05 Feb 2024 09:01:46 GMT",
            "host": "test_domain",
            "content-digest": "sha-256=:MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=:",
            "content-type": "application/activity+json",
            "signature": 'keyId="public_key_url",algorithm="rsa-sha256",headers="(request-target) host date content-digest content-type",signature="hUSgEIrqYzb0MOw48sNeKZdi18GEWOzKNSL1ZO/a+FOSfF6XWbCmMXDrVIco0nXbR+UtBoEKVH1w76j7MqsyuS0taUZLl2OlsOKwZqr8+etia6T+b6sQR+OTFpqQ+BO+MRFA5XCD+1AkydKyQvNWm68+oC4WF2aO9TDSnuKA5crUZfxngYIm+kt0s15JC3ahrHvkswQn26jA19cNoaTA0DlQiVPss9Q4AmLWfAQXHVprGQDqrlt5AmBZjk6iBd81d02GALZFOoQHt89pKEG0CEt3iTSyksAkPctDjsa6yAPgVOAdAx8pGrN2d86iQBz/d8Jg4tu5BGLbb760jcJw/Q=="',  # noqa F501
        },
        get_data=AsyncMock(return_value=b'{"cows": "good"}'),
    )

    assert await validate_http_signature(request)



# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from .multibase import multibase_58btc_encode


def test_multibase_58btc_encode():
    message = "secret".encode("utf-8")
    result = multibase_58btc_encode(message)

    assert result == "zzTuS2beK"



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from bovine.crypto import generate_ed25519_private_key, private_key_to_did_key


def main():
    private_key = generate_ed25519_private_key()
    did_key = private_key_to_did_key(private_key)

    print(f"Secret: {private_key}")
    print(f"Did-key: {did_key}")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import base64
import hashlib
import logging
import warnings
import jcs
from pyld import jsonld
from typing import Tuple, Union
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, padding, rsa, ec
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)
import based58
from http_sf import ser

from bovine.models import DataIntegrityProof
from bovine.utils import pydantic_to_json
from .multibase import (
    multibase_decode,
    encode_public_key_to_multibase,
    decode_multibase_public_key,
    MultiCodec,
)

logger = logging.getLogger(__name__)


def content_digest_sha256(content: str | bytes) -> str:
    """Computes the SHA256 digest of given content

    :param content: Content to be encoded"""
    if isinstance(content, str):
        content = content.encode("utf-8")

    digest = base64.standard_b64encode(hashlib.sha256(content).digest()).decode("utf-8")
    return "sha-256=" + digest


def content_digest_sha256_rfc_9530(content: bytes) -> Tuple[str, str]:
    """Computes the content digest according to [RFC 9530](https://www.rfc-editor.org/authors/rfc9530.html)

    ```pycon
    >>> content_digest_sha256_rfc_9530(b'{"hello": "world"}\\n')
    ('content-digest', 'sha-256=:RK/0qy18MlBSVnWgjwz6lZEWjP/lF5HF9bvEF8FabDg=:')

    ```

    :param content: Usually the request body to compute the digest from
    :return: Tuple of header name, header value"""
    return "content-digest", ser({"sha-256": hashlib.sha256(content).digest()})


def sign_message(private_key, message):
    warnings.warn("Deprecated will be removed with bovine 0.6.0", DeprecationWarning)
    try:
        key = load_pem_private_key(private_key.encode("utf-8"), password=None)
        assert isinstance(key, rsa.RSAPrivateKey)
    except Exception as e:
        logger.error(e)
        logger.error(private_key)
        raise (e)

    return base64.standard_b64encode(
        key.sign(
            message.encode("utf-8"),
            padding.PKCS1v15(),
            hashes.SHA256(),
        )
    ).decode("utf-8")


def verify_signature(public_key, message, signature):
    public_key_loaded = load_pem_public_key(public_key.encode("utf-8"))

    assert isinstance(public_key_loaded, rsa.RSAPublicKey)

    try:
        public_key_loaded.verify(
            base64.standard_b64decode(signature),
            message.encode("utf-8"),
            padding.PKCS1v15(),
            hashes.SHA256(),
        )
    except InvalidSignature:
        logger.warning("invalid signature")
        return False

    return True


def public_key_to_did_key(
    public_key: Union[
        ed25519.Ed25519PublicKey, rsa.RSAPublicKey, ec.EllipticCurvePublicKey
    ],
) -> str:
    """Converts a public key to a [did:key](https://w3c-ccg.github.io/did-method-key/).

    :param public_key: The public key
    """
    return "did:key:" + encode_public_key_to_multibase(public_key)


def private_key_to_base58(private_key: ed25519.Ed25519PrivateKey) -> str:
    private_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PrivateFormat.Raw,
        encryption_algorithm=serialization.NoEncryption(),
    )
    encoded = based58.b58encode(MultiCodec.Ed25519Private.value + private_bytes)
    return "z" + encoded.decode("ascii")


def did_key_to_public_key(did: str) -> ed25519.Ed25519PublicKey:
    if not did.startswith("did:key:"):
        raise ValueError(f"Should start with 'did:key:' got {str}")
    return decode_multibase_public_key(did[8:])


def jcs_sha256(doc: dict, context: dict | None = None) -> bytes:
    """Returns the sha256 digest of the representation
    of dict according to JCS. This assumes that `doc`
    is JSON serializable.

    JCS is defined in [RFC8785](https://www.rfc-editor.org/rfc/rfc8785).

    :param doc: The document
    :param context: ignored used in [rdfc_sha256][bovine.crypto.helper.rdfc_sha256]
    """
    return hashlib.sha256(jcs.canonicalize(doc)).digest()


def rdfc_sha256(doc: dict, context: dict = {}) -> bytes:
    """Returns the sha256 digest of the representation
    of dict according to [RDF-Canon](https://www.w3.org/TR/rdf-canon/)

    The algorithm used is URDNA2015 and the result are n-quads.

    :param doc: The document
    :param context: If doc has no `@context` property, it is set to this parameter
    """
    if "@context" not in doc:
        doc["@context"] = context

    normalized = jsonld.normalize(
        doc, {"algorithm": "URDNA2015", "format": "application/n-quads"}
    )
    return hashlib.sha256(normalized.encode("utf-8")).digest()


def jcs_sha384(doc: dict, context: dict | None = None) -> bytes:
    """Returns the sha384 digest of the representation
    of dict according to JCS. This assumes that `doc`
    is JSON serializable.

    JCS is defined in [RFC8785](https://www.rfc-editor.org/rfc/rfc8785).

    :param doc: The document
    :param context: ignored used in [rdfc_sha256][bovine.crypto.helper.rdfc_sha256]
    """
    return hashlib.sha384(jcs.canonicalize(doc)).digest()


def rdfc_sha384(doc: dict, context: dict = {}) -> bytes:
    """Returns the sha384 digest of the representation
    of dict according to [RDF-Canon](https://www.w3.org/TR/rdf-canon/)

    The algorithm used is URDNA2015 and the result are n-quads.

    :param doc: The document
    :param context: If doc has no `@context` property, it is set to this parameter
    """
    if "@context" not in doc:
        doc["@context"] = context

    normalized = jsonld.normalize(
        doc, {"algorithm": "URDNA2015", "format": "application/n-quads"}
    )
    logger.debug(normalized)
    return hashlib.sha384(normalized.encode("utf-8")).digest()


def ensure_data_integrity_context(doc: dict) -> dict:
    """Performs context injection according to
    [Verifiable Credentials: Data Integrity](https://w3c.github.io/vc-data-integrity/#context-injection). The data integrity
    context being added is `https://w3id.org/security/data-integrity/v2`.

    :param doc: Document
    :return: The same document with the data integrity context added.
    """
    integrity_context = "https://w3id.org/security/data-integrity/v2"
    alternate_context = "https://w3id.org/security/data-integrity/v1"
    context = doc.get("@context")
    if context is None:
        return {"@context": integrity_context, **doc}
    if isinstance(context, str):
        context: list = [context]
    if integrity_context in context or alternate_context in context:
        return doc
    return jsonld.compact(doc, context + [integrity_context])


def split_according_to_fep8b32(doc: dict) -> Tuple[dict, DataIntegrityProof, bytes]:
    pure_doc = {key: value for key, value in doc.items() if key != "proof"}
    proof_document = DataIntegrityProof.model_validate(doc["proof"])
    signature = multibase_decode(doc["proof"]["proofValue"])

    return pure_doc, proof_document, signature


def integrity_proof_to_pure_proof_dict(proof: DataIntegrityProof) -> dict:
    return {
        key: value
        for key, value in pydantic_to_json(proof).items()
        if key != "proofValue"
    }


def curve_to_hashing_algorithm_and_half_key_length(curve):
    if isinstance(curve, ec.SECP256R1):
        return ec.ECDSA(hashes.SHA256()), 32
    else:
        return ec.ECDSA(hashes.SHA384()), 48



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from . import OrderedCollectionPage


def test_ordered_collection_page() -> None:
    result = OrderedCollectionPage(
        id="url?page=1", part_of="url", items=["id1"]
    ).build()

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url?page=1",
        "partOf": "url",
        "orderedItems": ["id1"],
        "type": "OrderedCollectionPage",
    }

    result = OrderedCollectionPage(
        id="url?page=1", part_of="url", items=["id1"], next="next_url", prev="prev_url"
    ).build()

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url?page=1",
        "next": "next_url",
        "prev": "prev_url",
        "partOf": "url",
        "orderedItems": ["id1"],
        "type": "OrderedCollectionPage",
    }



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import logging
import warnings
from cryptography.exceptions import InvalidSignature

from .helper import (
    did_key_to_public_key,
    sign_message,
    verify_signature,
)

from .multibase import (
    multibase_decode,
    multibase_58btc_encode,
    multibase_to_private_key,
)

from .types import CryptographicSecret, CryptographicIdentifier

logger = logging.getLogger(__name__)


def build_signature(host, method, target):
    return (
        HttpSignature()
        .with_field("(request-target)", f"{method} {target}")
        .with_field("host", host)
    )


class HttpSignature:
    """Helper class to build http signatures

    Usage: Add fields used for signature with `with_fields`. Then
    use `build_signature` or `verify` depending on use case.
    """

    def __init__(self):
        self.fields = []

    def build_signature(self, key_id: str, private_key: str):
        """Returns the signature string when signed with private_key"""

        warnings.warn(
            "Deprecated use sign_for_http_draft instead, will be remove in bovine 0.6.0",
            DeprecationWarning,
        )

        message = self.build_message()

        signature_string = sign_message(private_key, message)
        headers = " ".join(name for name, _ in self.fields)

        signature_parts = [
            f'keyId="{key_id}"',
            'algorithm="rsa-sha256"',
            f'headers="{headers}"',
            f'signature="{signature_string}"',
        ]

        return ",".join(signature_parts)

    def sign_for_http_draft(self, secret: CryptographicSecret):
        message = self.build_message()

        signature_string = secret.sign(message)
        headers = " ".join(name for name, _ in self.fields)

        signature_parts = [
            f'keyId="{secret.key_id}"',
            'algorithm="rsa-sha256"',
            f'headers="{headers}"',
            f'signature="{signature_string}"',
        ]

        return ",".join(signature_parts)

    def ed25519_sign(self, private_encoded):
        private_key = multibase_to_private_key(private_encoded)

        message = self.build_message()

        return multibase_58btc_encode(private_key.sign(message.encode("utf-8")))

    def ed25519_verify(self, did_key, signature):
        public_key = did_key_to_public_key(did_key)

        if signature[0] != "z":
            raise ValueError(f"Expected signature to start with a z, got: {signature}")

        signature = multibase_decode(signature)

        message = self.build_message().encode("utf-8")

        try:
            public_key.verify(signature, message)
        except InvalidSignature:
            return False

        return True

    def verify(self, public_key: str, signature: str):
        """Verifies signature"""
        warnings.warn(
            "Deprecated use verify_with_identity instead, will be remove in bovine 0.6.0",
            DeprecationWarning,
        )

        message = self.build_message()
        return verify_signature(public_key, message, signature)

    def verify_with_identity(self, identifier: CryptographicIdentifier, signature: str):
        return identifier.verify(self.build_message(), signature)

    def build_message(self):
        """Builds the message"""
        return "\n".join(f"{name}: {value}" for name, value in self.fields)

    def with_field(self, field_name, field_value):
        """Adds a field to be used when building a http signature"""
        self.fields.append((field_name, field_value))
        return self

    @property
    def headers(self):
        """Headers as specified when building http signature"""
        return {name: value for name, value in self.fields if name[0] != "("}



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import json

from .helper import content_digest_sha256
from .http_signature import HttpSignature, build_signature
from .test import private_key


from .types import CryptographicSecret


def test_http_signature_sign_for_http_draft():
    http_signature = HttpSignature().with_field("name", "value")

    secret = CryptographicSecret.from_pem("key_id", private_key)
    signature_string = http_signature.sign_for_http_draft(secret)

    key_id, algorithm, headers, signature = signature_string.split(",")

    assert key_id == 'keyId="key_id"'
    assert algorithm == 'algorithm="rsa-sha256"'
    assert headers == 'headers="name"'
    assert (
        signature
        == 'signature="vpuy8U8A9w7Pr03956AaGg6VaLdZ6FTZuO6bsPL01SI1IIJj32mM/OCutktf0Fkqm1e1qcjDO6EVe3wZWRMjBN0GjE7NTj9wm8rBCQstdowVrwduHTuyPzWKzyoE+u7xBZPXwKy3uy1blC2BH4DJybZ94Hy2z+emrh3XKm+pi0xmMTdACvhizSP+inKHVbolFJ+zxj63wp/6AtkqzQEHbZq8X2DN1mIoDRMA4tPk0q9s5XMb02LyR1fNx5UeA2H/bijvgKY5OvstWbMXcUcPGVUA9elkPT23RNaCch63iaF1dmo/ke8JXy9zFZNx2UfmAO9yEclf6Azcz3kWy0SSQA=="'  # noqa F501
    )


def test_build_signature():
    http_signature = http_signature = build_signature(
        "myhost.tld", "get", "/path/to/resource"
    ).with_field("date", "Wed, 15 Mar 2023 17:28:15 GMT")

    secret = CryptographicSecret.from_pem("key_id", private_key)
    signature_string = http_signature.sign_for_http_draft(secret)

    assert (
        signature_string
        == 'keyId="key_id",algorithm="rsa-sha256",headers="(request-target) host date",signature="bwgA3UajpDzM07wM+MUkpGHS/mAhhTu+WDiBL1H28J762jUUuggzP8We5+I0WHypSBCWt8Lap2AbYqG6PfTsdFEbk8VWewX34n7/LUSM/WJeA8AFCEnHYPd8rRib4d0+kQV+U4Ai997XUxs2xi2AruNdkB6SNMkSbnQLoBjYT4tzGaf2I1xZuqjD1HvvW0xa1o/nPt4G5kj0tVmgDrrgeKL2hUJ21UdHcXDNFw9QJWf/O6KyEmbumvppKGFUfdpKSh3n2/fP7O4BxqadCbMYj0oDDOF1H3Pa6GV9EfIeSXXLS9TfyPpkXwbWrNpzUBZQ2hRGzX2jS/JEhpSNSVqv6A=="'  # noqa F501
    )

    key_id, algorithm, headers, signature = signature_string.split(",")

    assert key_id == 'keyId="key_id"'
    assert algorithm == 'algorithm="rsa-sha256"'
    assert headers == 'headers="(request-target) host date"'
    assert (
        signature
        == 'signature="bwgA3UajpDzM07wM+MUkpGHS/mAhhTu+WDiBL1H28J762jUUuggzP8We5+I0WHypSBCWt8Lap2AbYqG6PfTsdFEbk8VWewX34n7/LUSM/WJeA8AFCEnHYPd8rRib4d0+kQV+U4Ai997XUxs2xi2AruNdkB6SNMkSbnQLoBjYT4tzGaf2I1xZuqjD1HvvW0xa1o/nPt4G5kj0tVmgDrrgeKL2hUJ21UdHcXDNFw9QJWf/O6KyEmbumvppKGFUfdpKSh3n2/fP7O4BxqadCbMYj0oDDOF1H3Pa6GV9EfIeSXXLS9TfyPpkXwbWrNpzUBZQ2hRGzX2jS/JEhpSNSVqv6A=="'  # noqa F501
    )


def test_build_signature_post():
    http_signature = http_signature = (
        build_signature("myhost.tld", "post", "/path/to/resource")
        .with_field("date", "Wed, 15 Mar 2023 17:28:15 GMT")
        .with_field("digest", "sha-256=MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=")
    )

    secret = CryptographicSecret.from_pem("key_id", private_key)
    signature_string = http_signature.sign_for_http_draft(secret)

    assert (
        signature_string
        == 'keyId="key_id",algorithm="rsa-sha256",headers="(request-target) host date digest",signature="QpeLMEh/Z009DgYcSOkrsKr9zW7Wu7UWTQ398qvueCzaEL/Hxrv9C42U+WY/O34x/385lT+z1I3Bk6qMZTDAZSmKGYB5lZJVBdC20a4D++HzIyE62d6CvqUsQwIv4od/hOmmOQ3HKI69MUOsNrLyjIQxUaLSBz1m/wkLv8iMrh9QwULlqTO8WHVjSJlgVwllRtOOcm3wzo/cr2XyFj4gB8CUd3aG3dqzCfOOoNQvUhMDlxP7UqBnGoF6pnYnTnJUSSHdceWbE09TqXLmAHzauYGSdaHRS3XR/Sje7ET9UaOgHHcPJHIN+4jrBpDp2GM69joXWkZwhE0GOl9q3KBSKA=="'  # noqa F501
    )


def test_build_message():
    http_signature = build_signature(
        "myhost.tld", "get", "/path/to/resource"
    ).with_field("date", "Wed, 15 Mar 2023 17:28:15 GMT")

    message = http_signature.build_message()

    assert (
        message
        == """(request-target): get /path/to/resource
host: myhost.tld
date: Wed, 15 Mar 2023 17:28:15 GMT"""
    )

    signature = http_signature.ed25519_sign(
        "z3u2Yxcowsarethebestcowsarethebestcowsarethebest"
    )

    assert (
        signature
        == "z5ahdHCbP9aJEsDtvG1MEZpxPzuvGKYcdXdKvMq5YL21Z2"
        + "umxjs1SopCY2Ap8vZxVjTEf6dYbGuB7mtgcgUyNdBLe"
    )

    didkey = "did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"

    http_signature.ed25519_verify(didkey, signature)


def test_build_message_post():
    body = json.dumps({"cows": "good"}).encode("utf-8")

    assert body == b'{"cows": "good"}'

    digest = content_digest_sha256(body)

    assert digest == "sha-256=MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k="

    http_signature = (
        build_signature("myhost.tld", "post", "/path/to/resource")
        .with_field("date", "Wed, 15 Mar 2023 17:28:15 GMT")
        .with_field("digest", digest)
    )

    message = http_signature.build_message()

    assert (
        message
        == """(request-target): post /path/to/resource
host: myhost.tld
date: Wed, 15 Mar 2023 17:28:15 GMT
digest: sha-256=MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k="""
    )

    signature = http_signature.ed25519_sign(
        "z3u2Yxcowsarethebestcowsarethebestcowsarethebest"
    )

    assert (
        signature
        == "z4vPkJaoaSVQp5DrMb8EvCajJcerW36rsyWDELTWQ3cYmaonnGf"
        + "b8WHiwH54BShidCcmpoyHjanVRYNrXXXka4jAn"
    )

    didkey = "did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"

    http_signature.ed25519_verify(didkey, signature)



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from typing import Dict

import aiohttp

from bovine.types import ServerSentEvent


class EventSource:
    def __init__(
        self, session: aiohttp.ClientSession, url: str, headers: Dict[str, str] = {}
    ):
        self.session = session
        self.url: str = url
        self.headers = headers
        self.response: aiohttp.ClientResponse | None = None

    async def create_response(self) -> None:
        timeout = aiohttp.ClientTimeout(total=None)

        self.response = await self.session.get(
            self.url,
            headers={"Accept": "text/event-stream", **self.headers},
            timeout=timeout,
            allow_redirects=False,
        )
        assert self.response
        self.response.raise_for_status()

        if self.response.status == 301:
            self.url = self.response.headers["location"]
            await self.create_response()

    def __aiter__(self):
        return self

    async def __anext__(self) -> ServerSentEvent | None:
        if self.response is None:
            await self.create_response()

        to_parse = ""

        assert self.response

        async for line_in_bytes in self.response.content:
            line = line_in_bytes.decode("utf-8")
            if line[0] == ":":
                continue
            if line == "\n":
                event = ServerSentEvent.parse_utf8(to_parse)
                if event:
                    return event
            else:
                to_parse = f"{to_parse}{line}"

        return None



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from .multikey import Multikey


def test_multikey_from_multibase():
    multikey = Multikey(
        id="https://server.example/users/alice#ed25519-key",
        controller="https://server.example/users/alice",
        multibase="z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2",
    )
    expected = {
        "@context": "https://w3id.org/security/multikey/v1",
        "id": "https://server.example/users/alice#ed25519-key",
        "type": "Multikey",
        "controller": "https://server.example/users/alice",
        "publicKeyMultibase": "z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2",
    }

    assert multikey.build() == expected


def test_from_multibase_and_controller():
    multikey = Multikey.from_multibase_and_controller(
        "https://server.example/users/alice",
        "z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2",
    )
    expected = {
        "@context": "https://w3id.org/security/multikey/v1",
        "id": "https://server.example/users/alice#z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2",
        "type": "Multikey",
        "controller": "https://server.example/users/alice",
        "publicKeyMultibase": "z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2",
    }

    assert multikey.build() == expected



# SPDX-FileCopyrightText: 2023-2024 Helge
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass


@dataclass
class Multikey:
    """Represents a Multikey"""

    id: str
    controller: str
    multibase: str
    type: str = "Multikey"

    def build(self):
        return {
            "@context": "https://w3id.org/security/multikey/v1",
            "id": self.id,
            "type": self.type,
            "controller": self.controller,
            "publicKeyMultibase": self.multibase,
        }

    @staticmethod
    def from_multibase_and_controller(controller, multibase):
        return Multikey(
            id=f"{controller}#{multibase}", controller=controller, multibase=multibase
        )



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import logging
import os
import traceback
from typing import Optional, Tuple

import aiodns
import aiohttp

from bovine.utils import parse_fediverse_handle

from .lookup_account import lookup_with_webfinger
from .nodeinfo import fetch_nodeinfo20, fetch_nodeinfo_document

logger = logging.getLogger(__name__)


async def lookup_account_with_webfinger(
    session: aiohttp.ClientSession, fediverse_handle: str
) -> str | None:
    """**Deprecated**: Use lookup_uri_with_webfinger instead

    Looks up the actor url associated with a FediVerse handle,
    i.e. an identifier of the form username@domain, using
    the webfinger endpoint

    :param session: the aiohttp.ClientSession to use
    :param fediverse_handle: the FediVerse handle as a string
    """
    result, _ = await lookup_uri_with_webfinger(session, "acct:" + fediverse_handle)
    return result


async def lookup_did_with_webfinger(
    session: aiohttp.ClientSession, domain: str, did: str
):
    """**Deprecated**: Use lookup_uri_with_webfinger instead

    Looks up the actor url associated with a did and domain
    using the webfinger endpoint

    :param session: the aiohttp.ClientSession to use
    :param domain: the domain to perform the lookup from
    :param did: the did key to perform lookup with
    """
    result, _ = await lookup_uri_with_webfinger(session, did, domain=domain)
    return result


async def lookup_uri_with_webfinger(
    session: aiohttp.ClientSession, uri: str, domain: str | None = None
) -> Tuple[Optional[str], Optional[bool]]:
    """Looks up an actor URI associated with an URI and domain
    using the webfinger endpoint following `fep-4adb
    <https://codeberg.org/fediverse/fep/src/branch/main/feps/fep-4adb.md>`_

    :param session: the aiohttp.ClientSession to use
    :param uri: the uri to perform lookup with
    :param domain: the domain to perform the lookup from
    :returns: A tuple of `(actor URI, verification_necessary)`. The value
        of `verification_necessary` indicates that it is necessary if the actor
        has the appropriate authority to associate the query URI with itself.

    """
    if domain is None:
        if uri.startswith("acct:"):
            _, domain = parse_fediverse_handle(uri.removeprefix("acct:"))
        else:
            raise ValueError(f"For the uri {uri} a domain must be specified")

    if domain.startswith("http://") or domain.startswith("https://"):
        webfinger_url = f"{domain}/.well-known/webfinger"
    else:
        webfinger_url = f"https://{domain}/.well-known/webfinger"
    params = {"resource": uri}

    try:
        result = await lookup_with_webfinger(session, webfinger_url, params)
    except aiohttp.client_exceptions.ClientConnectorError as ex:
        if os.environ.get("BUTCHER_ALLOW_HTTP"):
            webfinger_url = webfinger_url.replace("https://", "http://")
            result = await lookup_with_webfinger(session, webfinger_url, params)
        else:
            raise ex
    return result, not uri.startswith("acct:")


async def lookup_with_dns(session: aiohttp.ClientSession, domain: str) -> str | None:
    """Looks up the actor url associated with the dns entry for domain. See
    `FEP-612d: Identifying ActivityPub Objects through DNS
    <https://codeberg.org/fediverse/fep/src/branch/main/feps/fep-612d.md>`_ for
    the mechanism.

    :param session: the aiohttp.ClientSession to use
    :param domain: the domain to perform the lookup from
    """

    resolver = aiodns.DNSResolver()
    try:
        (result,) = await resolver.query(f"_apobjid.{domain}", "TXT")

        return result.text
    except Exception:
        return None


async def fetch_nodeinfo(session: aiohttp.ClientSession, domain: str) -> dict | None:
    """Fetches the nodeinfo 2.0 object from domain using the /.well-known/nodeinfo
    endpoint"""

    try:
        data = await fetch_nodeinfo_document(session, domain)

        for link in data["links"]:
            if link["rel"] == "http://nodeinfo.diaspora.software/ns/schema/2.0":
                return await fetch_nodeinfo20(session, link["href"])

        return None

    except Exception as e:
        logger.error(str(e))
        for log_line in traceback.format_exc().splitlines():
            logger.error(log_line)
        return None



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

import pytest

from .types import CryptographicIdentifier


def test_verify_document():
    credential = {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            "https://w3id.org/security/data-integrity/v2",
        ],
        "id": "urn:uuid:86294362-4254-4f36-854f-3952fe42555d",
        "type": ["VerifiableCredential"],
        "issuer": "did:key:z6MktKwz7Ge1Yxzr4JHavN33wiwa8y81QdcMRLXQsrH9T53b",
        "issuanceDate": "2020-03-16T22:37:26.544Z",
        "credentialSubject": {
            "id": "did:key:z6MktKwz7Ge1Yxzr4JHavN33wiwa8y81QdcMRLXQsrH9T53b"
        },
        "proof": {
            "type": "DataIntegrityProof",
            "created": "2023-12-28T11:26:24Z",
            "verificationMethod": "did:key:z6MktKwz7Ge1Yxzr4JHavN33wiwa8y81QdcMRLXQsrH9T53b#z6MktKwz7Ge1Yxzr4JHavN33wiwa8y81QdcMRLXQsrH9T53b",
            "cryptosuite": "eddsa-rdfc-2022",
            "proofPurpose": "assertionMethod",
            "proofValue": "z59CRYvhTxnciJNtNAhp4fb5CbQvkMAZywYPrxDNu2d7CLn2kemf52BfnpA9hJJejzUv2zEmAA545ZBcf4aPg2BaM",
        },
    }
    did_key = "did:key:z6MktKwz7Ge1Yxzr4JHavN33wiwa8y81QdcMRLXQsrH9T53b"
    identifier = CryptographicIdentifier.from_did_key(did_key)

    assert identifier.verify_document(credential) == did_key


# Test data from https://www.w3.org/TR/vc-di-ecdsa/#representation-ecdsa-rdfc-2019-with-curve-p-256


def test_verify_document_w3c_recommendation_p256():
    credential = {
        "@context": [
            "https://www.w3.org/ns/credentials/v2",
            "https://www.w3.org/ns/credentials/examples/v2",
        ],
        "id": "urn:uuid:58172aac-d8ba-11ed-83dd-0b3aef56cc33",
        "type": ["VerifiableCredential", "AlumniCredential"],
        "name": "Alumni Credential",
        "description": "A minimum viable example of an Alumni Credential.",
        "issuer": "https://vc.example/issuers/5678",
        "validFrom": "2023-01-01T00:00:00Z",
        "credentialSubject": {
            "id": "did:example:abcdefgh",
            "alumniOf": "The School of Examples",
        },
        "proof": {
            "type": "DataIntegrityProof",
            "cryptosuite": "ecdsa-rdfc-2019",
            "created": "2023-02-24T23:36:38Z",
            "verificationMethod": "https://vc.example/issuers/5678#zDnaepBuvsQ8cpsWrVKw8fbpGpvPeNSjVPTWoq6cRqaYzBKVP",
            "proofPurpose": "assertionMethod",
            "proofValue": "z4KKHqaD4F7GHyLA6f3wK9Ehxtogv5jQRFpQBM4sXkSf7Bozd7bAf7dF6UkfM2aSCBMm24mPvaFXmzQmimzaEC3SL",
        },
    }
    controller = "https://vc.example/issuers/5678"

    identifier = CryptographicIdentifier.from_tuple(
        controller, "zDnaepBuvsQ8cpsWrVKw8fbpGpvPeNSjVPTWoq6cRqaYzBKVP"
    )
    assert identifier.verify_document(credential) == controller


# Test data from https://www.w3.org/TR/vc-di-ecdsa/#representation-ecdsa-rdfc-2019-with-curve-p-384


def test_verify_document_w3c_recommendation_p386():
    credential = {
        "@context": [
            "https://www.w3.org/ns/credentials/v2",
            "https://www.w3.org/ns/credentials/examples/v2",
        ],
        "id": "urn:uuid:58172aac-d8ba-11ed-83dd-0b3aef56cc33",
        "type": ["VerifiableCredential", "AlumniCredential"],
        "name": "Alumni Credential",
        "description": "A minimum viable example of an Alumni Credential.",
        "issuer": "https://vc.example/issuers/5678",
        "validFrom": "2023-01-01T00:00:00Z",
        "credentialSubject": {
            "id": "did:example:abcdefgh",
            "alumniOf": "The School of Examples",
        },
        "proof": {
            "type": "DataIntegrityProof",
            "cryptosuite": "ecdsa-rdfc-2019",
            "created": "2023-02-24T23:36:38Z",
            "verificationMethod": "https://vc.example/issuers/5678#z82LkuBieyGShVBhvtE2zoiD6Kma4tJGFtkAhxR5pfkp5QPw4LutoYWhvQCnGjdVn14kujQ",
            "proofPurpose": "assertionMethod",
            "proofValue": "zpuEu1cJ7Wpb453b4RiV3ex7SKGYm3fdAd4WUTVpR8Me3ZXkCCVUfd4M4TvHF9Wv1tRNWe5SkZhQTGYLUxdugFRGC2uyYRNTnimS6UMN6wkenTViRK1Mei7DooSBpumHHjYu",
        },
    }
    controller = "https://vc.example/issuers/5678"

    identifier = CryptographicIdentifier.from_tuple(
        controller,
        "z82LkuBieyGShVBhvtE2zoiD6Kma4tJGFtkAhxR5pfkp5QPw4LutoYWhvQCnGjdVn14kujQ",
    )
    assert identifier.verify_document(credential) == controller


# Test data from https://www.w3.org/TR/vc-di-ecdsa/#representation-ecdsa-jcs-2019-with-curve-p-384


def test_verify_document_w3c_recommendation_p386_jcs():
    credential = {
        "@context": [
            "https://www.w3.org/ns/credentials/v2",
            "https://www.w3.org/ns/credentials/examples/v2",
        ],
        "id": "urn:uuid:58172aac-d8ba-11ed-83dd-0b3aef56cc33",
        "type": ["VerifiableCredential", "AlumniCredential"],
        "name": "Alumni Credential",
        "description": "A minimum viable example of an Alumni Credential.",
        "issuer": "https://vc.example/issuers/5678",
        "validFrom": "2023-01-01T00:00:00Z",
        "credentialSubject": {
            "id": "did:example:abcdefgh",
            "alumniOf": "The School of Examples",
        },
        "proof": {
            "type": "DataIntegrityProof",
            "cryptosuite": "ecdsa-jcs-2019",
            "created": "2023-02-24T23:36:38Z",
            "verificationMethod": "https://vc.example/issuers/5678#z82LkuBieyGShVBhvtE2zoiD6Kma4tJGFtkAhxR5pfkp5QPw4LutoYWhvQCnGjdVn14kujQ",
            "proofPurpose": "assertionMethod",
            "proofValue": "zFYhRwKuucKxM7dnL69VpnwmU9UD2wc5HfFjXfxKH82pEybv18EfxaT8m53kyMfrDQneYnsLCZ35UE2KwZTkd4zN7vNHdVseyjW5apJJ9NkfpUiTGUayG2yaZvWu6Gd8EDYk",
        },
    }

    controller = "https://vc.example/issuers/5678"

    identifier = CryptographicIdentifier.from_tuple(
        controller,
        "z82LkuBieyGShVBhvtE2zoiD6Kma4tJGFtkAhxR5pfkp5QPw4LutoYWhvQCnGjdVn14kujQ",
    )
    assert identifier.verify_document(credential) == controller


# credential from https://github.com/w3c/vc-test-suite-implementations/pull/11#issuecomment-1874285663


def test_verify_document_from_digital_bazaar():
    credential = {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            {
                "@protected": True,
                "DriverLicenseCredential": "urn:example:DriverLicenseCredential",
                "DriverLicense": {
                    "@id": "urn:example:DriverLicense",
                    "@context": {
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "documentIdentifier": "urn:example:documentIdentifier",
                        "dateOfBirth": "urn:example:dateOfBirth",
                        "expirationDate": "urn:example:expiration",
                        "issuingAuthority": "urn:example:issuingAuthority",
                    },
                },
                "driverLicense": {"@id": "urn:example:driverLicense", "@type": "@id"},
            },
            "https://w3id.org/security/data-integrity/v2",
        ],
        "id": "urn:uuid:6a6a9bb2-090a-4ffa-8462-4a614ae4c269",
        "type": ["VerifiableCredential", "DriverLicenseCredential"],
        "credentialSubject": {
            "id": "urn:uuid:1a0e4ef5-091f-4060-842e-18e519ab9440",
            "driverLicense": {
                "type": "DriverLicense",
                "documentIdentifier": "T21387yc328c7y32h23f23",
                "dateOfBirth": "01-01-1990",
                "expirationDate": "01-01-2030",
                "issuingAuthority": "VA",
            },
        },
        "issuer": "did:key:zDnaeRZqxu4dSRPiLWwph8ghUy17tE9BWXTNqtpzTw2HjthdX",
        "issuanceDate": "2024-01-02T16:13:22Z",
        "proof": {
            "type": "DataIntegrityProof",
            "created": "2024-01-02T16:13:22Z",
            "verificationMethod": "did:key:zDnaeRZqxu4dSRPiLWwph8ghUy17tE9BWXTNqtpzTw2HjthdX#zDnaeRZqxu4dSRPiLWwph8ghUy17tE9BWXTNqtpzTw2HjthdX",
            "cryptosuite": "ecdsa-rdfc-2019",
            "proofPurpose": "assertionMethod",
            "proofValue": "z65xxPTTzV6w4Jtjiscbr5cX6QT5XJFZKC74hE6pzeboEGXknnVu8qT3JhBAJ12MdBWrVsEsuqWbh9Mb8StgF563N",
        },
    }
    did_key = "did:key:zDnaeRZqxu4dSRPiLWwph8ghUy17tE9BWXTNqtpzTw2HjthdX"
    identifier = CryptographicIdentifier.from_did_key(did_key)

    assert identifier.verify_document(credential) == did_key


# credential from https://github.com/w3c/vc-test-suite-implementations/pull/11#issuecomment-1874285663


@pytest.mark.skip
def test_verify_document_from_digital_bazaar_p386():
    credential = {
        "@context": [
            "https://www.w3.org/2018/credentials/v1",
            {
                "@protected": True,
                "DriverLicenseCredential": "urn:example:DriverLicenseCredential",
                "DriverLicense": {
                    "@id": "urn:example:DriverLicense",
                    "@context": {
                        "@protected": True,
                        "id": "@id",
                        "type": "@type",
                        "documentIdentifier": "urn:example:documentIdentifier",
                        "dateOfBirth": "urn:example:dateOfBirth",
                        "expirationDate": "urn:example:expiration",
                        "issuingAuthority": "urn:example:issuingAuthority",
                    },
                },
                "driverLicense": {"@id": "urn:example:driverLicense", "@type": "@id"},
            },
            "https://w3id.org/security/data-integrity/v2",
        ],
        "id": "urn:uuid:1e30cef5-6a61-4703-809f-9b7df5bde1a3",
        "type": ["VerifiableCredential", "DriverLicenseCredential"],
        "credentialSubject": {
            "id": "urn:uuid:1a0e4ef5-091f-4060-842e-18e519ab9440",
            "driverLicense": {
                "type": "DriverLicense",
                "documentIdentifier": "T21387yc328c7y32h23f23",
                "dateOfBirth": "01-01-1990",
                "expirationDate": "01-01-2030",
                "issuingAuthority": "VA",
            },
        },
        "issuer": "did:key:z82Lky21MA8mGLTpsgvGnikRjS12Ar8wik7L3BJi9xSvU89rZaKcNRrcddPHhNFTWdR84oB",
        "issuanceDate": "2024-01-03T15:15:20Z",
        "proof": {
            "type": "DataIntegrityProof",
            "created": "2024-01-03T15:15:20Z",
            "verificationMethod": "did:key:z82Lky21MA8mGLTpsgvGnikRjS12Ar8wik7L3BJi9xSvU89rZaKcNRrcddPHhNFTWdR84oB#z82Lky21MA8mGLTpsgvGnikRjS12Ar8wik7L3BJi9xSvU89rZaKcNRrcddPHhNFTWdR84oB",
            "cryptosuite": "ecdsa-rdfc-2019",
            "proofPurpose": "assertionMethod",
            "proofValue": "zrqP9MFHdZuRfF8jeFm57dQsqGpsv2JNHjCUy1Vrm7QJqwyesYkVv3ewn2SyH6MMgvVWzhNUziBUwZtwGW7CkA81GTMPYBb2ubKvMTXqhq2kwVwbC1oSEECevDHuhxdGqHaY",
        },
    }

    did_key = "did:key:z82Lky21MA8mGLTpsgvGnikRjS12Ar8wik7L3BJi9xSvU89rZaKcNRrcddPHhNFTWdR84oB"
    identifier = CryptographicIdentifier.from_did_key(did_key)

    assert identifier.verify_document(credential) == did_key


def test_verify_document_unsupported_cryptosuite():
    credential = {
        "@context": ["https://www.w3.org/2018/credentials/v1"],
        "id": "urn:uuid:09327695-56d3-4ef2-890c-d63b2b5e707f",
        "type": ["VerifiableCredential"],
        "credentialSubject": {
            "id": "did:key:z6MktKwz7Ge1Yxzr4JHavN33wiwa8y81QdcMRLXQsrH9T53b"
        },
        "issuer": "did:key:z6MkgND5U5Kedizov5nxeh2ZCVUTDRSmAfbNqPhzCq8b72Ra",
        "issuanceDate": "2020-03-16T22:37:26.544Z",
        "proof": {
            "@context": ["https://w3id.org/security/data-integrity/v1"],
            "type": "DataIntegrityProof",
            "proofPurpose": "assertionMethod",
            "proofValue": "z4ncyz44KMkma4Z4h3iVRx6W2tB7xVbMNGk8gc9BRtqNWWsixNazdshm1aZADEH3vNJxd9RMKBTMqhSCcSEH6tn8V",
            "verificationMethod": "did:key:z6MkgND5U5Kedizov5nxeh2ZCVUTDRSmAfbNqPhzCq8b72Ra#z6MkgND5U5Kedizov5nxeh2ZCVUTDRSmAfbNqPhzCq8b72Ra",
            "created": "2024-02-05T10:21:42.315Z",
            "cryptosuite": "eddsa-2022",
        },
    }

    did_key = "did:key:z6MkgND5U5Kedizov5nxeh2ZCVUTDRSmAfbNqPhzCq8b72Ra"
    identifier = CryptographicIdentifier.from_did_key(did_key)

    with pytest.raises(ValueError) as info:
        identifier.verify_document(credential)

    assert str(info.value) == "Unknown cryptosuite 'eddsa-2022'"



# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import pytest

from .cache import JsonLDCache


@pytest.fixture
def cache():
    yield JsonLDCache(filename=":memory:")


def test_creation(cache):
    assert cache.retrieve("https://one.test/") is None


def test_add_element(cache):
    context = {"hello": "world"}

    cache.add("https://two.test/", context)

    assert cache.retrieve("https://one.test/") is None
    assert cache.retrieve("https://two.test/") == context


def test_load(tmp_path):
    filename = tmp_path / "cache.sqlite"
    cache = JsonLDCache(filename=filename)
    context = {"hello": "world"}
    cache.add("https://two.test/", context)

    loaded = JsonLDCache(filename=filename)
    assert loaded.retrieve("https://two.test/") == context



# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from enum import Enum
from typing import Union

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, rsa, ec
from cryptography.hazmat.primitives.serialization import load_der_public_key
import based58


class MultiCodec(Enum):
    """The used Multicodec prefixes"""

    Ed25519Public = b"\xed\x01"
    EcP256Public = b"\x80$"
    EcP384Public = b"\x81\x24"
    RsaPublic = b"\x85$"

    Ed25519Private = b"\x80\x26"
    EcP256Private = b"\x86\x26"


def multibase_58btc_encode(data: bytes) -> str:
    """Encodes `data` in base 58 using the bitcoin alphabet
    and adds the prefix `z`"""
    return "z" + based58.b58encode(data).decode("utf-8")


def multibase_decode(data: str) -> bytes:
    """Decodes the string data using the multibase algorithm

    :param data: The string to decode
    :return: The bytes"""
    if data[0] == "z":
        return based58.b58decode(data[1:].encode("utf-8"))

    raise ValueError(f"{data} encoded in unknown format")


def encode_public_key_to_multibase(public_key) -> str:
    match public_key:
        case rsa.RSAPublicKey():
            public_bytes = MultiCodec.RsaPublic.value + public_key.public_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
            return multibase_58btc_encode(public_bytes)
        case ed25519.Ed25519PublicKey():
            return multibase_58btc_encode(
                MultiCodec.Ed25519Public.value
                + public_key.public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw,
                )
            )
        case ec.EllipticCurvePublicKey():
            public_bytes = public_key.public_bytes(
                serialization.Encoding.X962,
                serialization.PublicFormat.CompressedPoint,
            )
            if isinstance(public_key.curve, ec.SECP256R1):
                return multibase_58btc_encode(
                    MultiCodec.EcP256Public.value + public_bytes
                )
            if isinstance(public_key.curve, ec.SECP384R1):
                return multibase_58btc_encode(
                    MultiCodec.EcP384Public.value + public_bytes
                )
    raise ValueError("Unsupported key type")


def decode_multibase_public_key(multibase_public_key):
    public_key = multibase_decode(multibase_public_key)

    match public_key[:2]:
        case MultiCodec.Ed25519Public.value:
            return ed25519.Ed25519PublicKey.from_public_bytes(public_key[2:])
        case MultiCodec.RsaPublic.value:
            return load_der_public_key(public_key[2:])
        case MultiCodec.EcP256Public.value:
            return ec.EllipticCurvePublicKey.from_encoded_point(
                ec.SECP256R1(), public_key[2:]
            )
        case MultiCodec.EcP384Public.value:
            return ec.EllipticCurvePublicKey.from_encoded_point(
                ec.SECP384R1(), public_key[2:]
            )
    return None


def multibase_to_private_key(
    multibase: str,
) -> Union[ed25519.Ed25519PrivateKey, ec.EllipticCurvePrivateKey]:
    decoded = multibase_decode(multibase)
    if decoded[:2] == MultiCodec.Ed25519Private.value:
        return ed25519.Ed25519PrivateKey.from_private_bytes(decoded[2:])
    if decoded[:2] == MultiCodec.EcP256Private.value:
        return ec.derive_private_key(int.from_bytes(decoded[2:]), ec.SECP256R1())
    raise ValueError(f"Improper start for ed25519 private key. Got {str}")



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import logging
import re
from dataclasses import dataclass
from enum import Enum

logger = logging.getLogger(__name__)


class Visibility(Enum):
    """Defines the visiblity of an object. Currently only used for Actor,
    where endpoints are hidden from anyone by the owner."""

    WEB = "WEB"
    PUBLIC = "PUBLIC"
    RESTRICTED = "RESTRICTED"
    OWNER = "OWNER"


sse_line_pattern = re.compile("(?P<name>[^:]*):?( ?(?P<value>.*))?")


@dataclass
class ServerSentEvent:
    """A server sent event as defined in
    `Event Source Interface
    <https://html.spec.whatwg.org/multipage/server-sent-events.html#server-sent-events>`_.
    This is used by the event_source function
    of the BovineClient."""

    data: str
    event: str | None = None
    id: str | None = None
    retry: int | None = None

    def encode(self) -> bytes:
        """Encodes the server sent event as bytes to send to the client."""
        message = f"data: {self.data}"
        if self.event is not None:
            message = f"{message}\nevent: {self.event}"
        if self.id is not None:
            message = f"{message}\nid: {self.id}"
        if self.retry is not None:
            message = f"{message}\nretry: {self.retry}"
        message = f"{message}\n\n"
        return message.encode("utf-8")

    @staticmethod
    def parse(raw: bytes):
        """Parses bytes to ServerSentEvent"""
        return ServerSentEvent.parse_utf8(raw.decode("utf-8"))

    @staticmethod
    def parse_utf8(raw: str):
        """Parses utf-8 encoded string to ServerSentEvent"""
        data = None
        event = None
        event_id = None
        retry = None
        for line in raw.splitlines():
            m = sse_line_pattern.match(line)
            if m is None:
                logger.warning(f"Invalid Server Sent Event line: '{line}'")
                continue

            name = m.group("name")
            if name == "":
                continue

            value = m.group("value")

            if name == "data":
                if data:
                    data = f"{data}\n{value}"
                else:
                    data = value
            elif name == "event":
                event = value
            elif name == "id":
                event_id = value
            elif name == "retry":
                retry = int(value)

        if not isinstance(data, str):
            return None

        return ServerSentEvent(data, event, event_id, retry)



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import pytest
from cryptography.hazmat.primitives.asymmetric import ed25519, rsa, ec

from .types import CryptographicIdentifier, CryptographicSecret
from .test import public_key, did_key, private_key, ed25519_key, ecp256_key


@pytest.mark.parametrize(
    "example",
    [
        {},
        {"type": "Key"},
        {"type": "Multikey"},
        {"type": "Multikey", "controller": "controller"},
        {"type": "Multikey", "publicKeyMultibase": "zxxx"},
        {"type": "Multikey", "controller": "controller", "publicKeyMultibase": "zxxx"},
        {
            "type": "Key",
            "controller": "controller",
            "publicKeyMultibase": did_key.removeprefix("did:key:"),
        },
    ],
)
def test_cryptographic_identity_from_multikey_error(example):
    with pytest.raises(ValueError):
        CryptographicIdentifier.from_multikey(example)


def test_cryptographic_identity_from_multikey():
    public_key_ed25519 = did_key.removeprefix("did:key:")
    controller = "https://com.example/issuer/123"
    multikey = {
        "id": f"{controller}#key-0",
        "type": "Multikey",
        "controller": controller,
        "publicKeyMultibase": public_key_ed25519,
    }

    identity = CryptographicIdentifier.from_multikey(multikey)

    assert identity.controller == controller

    signature = "0gHDz8Qn4+8gWC/byjTk7yvJxL0p4kiUSdxt6VZQvWQ9MBlRThMBDSrgJsNPHZWNMXtPSoL9+r0k\n3cUwYmIWDA=="

    assert identity.verify("moo", signature)
    assert identity.as_tuple() == (controller, public_key_ed25519)


def test_cryptographic_identity_from_multikey_rsa():
    controller = "https://com.example/issuer/123"
    public_key_rsa = "z4MXj1wBzi9jUstyQ1t9N9BHZpoQb4FQpaUfcc61XQigH9ua5R9aEN7YMK81PTKgWLGZXMqvy9eP4X42KgEsmNnZrwQbeT5R8oKhMHVQ5qcwTFwdX2gQeZLLDkUDJL3aJqmqSLK7mrgZ1CskMyD4p8eqFEUW1oufy9cE6Wyz7TQZFKpSCd1oY8HNue9cNRthZzXCdoX6DGVyewBFdivkohE1mhU1EpbKSYH66rx1cZpa6PJKzg4LbKSUqhHaftmsD1jWzFrKNUFzRmCGsihAjLVgsfjPaPmBUXNjYTFg1nCHWCGVGD3g9NhBwGiuu4vQR5PQfD6BCPZpGTaUZjWgZTHveef1pUDPvsCRuGDoGvrTnG8k7SeQp"
    multikey = {
        "id": f"{controller}#key-0",
        "type": "Multikey",
        "controller": controller,
        "publicKeyMultibase": public_key_rsa,
    }

    identity = CryptographicIdentifier.from_multikey(multikey)

    assert identity.controller == controller
    assert identity.as_tuple()[0] == controller

    identity_too = CryptographicIdentifier.from_tuple(*identity.as_tuple())

    assert identity.public_key == identity_too.public_key


@pytest.mark.parametrize(
    "example",
    [
        {},
        {"owner": "owner"},
        {"publicKeyPem": "xxxx"},
        {"owner": "owner", "publicKeyPem": "xxxx"},
    ],
)
def test_cryptographic_identity_from_public_key_error(example):
    with pytest.raises(ValueError):
        CryptographicIdentifier.from_public_key(example)


def test_cryptographic_identity_from_public_key():
    controller = "https://com.example/issuer/123"

    public_key_dict = {
        "id": f"{controller}/main-key",
        "owner": controller,
        "publicKeyPem": public_key,
    }

    identity = CryptographicIdentifier.from_public_key(public_key_dict)

    assert identity.controller == controller

    signature = "vaSYmwpEhGhB/o5QNC8MxYbJeBKDDiaZG0J4EsN0V/5+bFRgPbFK1oUgrdkiTT+farWXdVagPvIg44M/IYjPY8oExBS3mCt9oXDDWiDfBED8n2yrGHV6X/GWNxWUarmo4RcOBU2xWy9982/ZH+UyiPVEpanPi4REf9UYiF0dciZK1Yx3Nkqadnm9XTJJISHX4v88jkUGYaNnWcJ+SJMXuqklYJU/j8j4FVvf3vbFvuwGX1W5o7Zmk89xdJeRiGPYCM2zUgfzGoDHdVHuX8ksR+/xjzwLxMn/SerHuKSCzYivCpaxUqmX0VMTEwvPZ2H+hvXsqyLgR+zFnL7WdX6p7Q=="

    assert identity.verify("secret", signature)
    assert not identity.verify("secret", "")


@pytest.mark.parametrize(
    "example",
    ["", "zxxxxx", "did:key:zxdsafh"],
)
def test_cryptographic_identity_from_did_key_error(example):
    with pytest.raises(ValueError):
        CryptographicIdentifier.from_did_key(example)


# Some examples for this tests are from https://w3c-ccg.github.io/did-method-key
@pytest.mark.parametrize(
    ["did", "key_type"],
    [
        (
            "did:key:z6MknGc3ocHs3zdPiJbnaaqDi58NGb4pk1Sp9WxWufuXSdxf",
            ed25519.Ed25519PublicKey,
        ),
        (
            "did:key:z6MkiTBz1ymuepAQ4HEHYSF1H8quG5GLVVQR3djdX3mDooWp",
            ed25519.Ed25519PublicKey,
        ),
        (
            "did:key:z4MXj1wBzi9jUstyPMS4jQqB6KdJaiatPkAtVtGc6bQEQEEsKTic4G7Rou3iBf9vPmT5dbkm9qsZsuVNjq8HCuW1w24nhBFGkRE4cd2Uf2tfrB3N7h4mnyPp1BF3ZttHTYv3DLUPi1zMdkULiow3M1GfXkoC6DoxDUm1jmN6GBj22SjVsr6dxezRVQc7aj9TxE7JLbMH1wh5X3kA58H3DFW8rnYMakFGbca5CB2Jf6CnGQZmL7o5uJAdTwXfy2iiiyPxXEGerMhHwhjTA1mKYobyk2CpeEcmvynADfNZ5MBvcCS7m3XkFCMNUYBS9NQ3fze6vMSUPsNa6GVYmKx2x6JrdEjCk3qRMMmyjnjCMfR4pXbRMZa3i",
            rsa.RSAPublicKey,
        ),
        (
            "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169",
            ec.EllipticCurvePublicKey,
        ),
        (
            "did:key:zDnaerx9CtbPJ1q36T5Ln5wYt3MQYeGRG5ehnPAmxcf5mDZpv",
            ec.EllipticCurvePublicKey,
        ),
        (
            "did:key:z82Lm1MpAkeJcix9K8TMiLd5NMAhnwkjjCBeWHXyu3U4oT2MVJJKXkcVBgjGhnLBn2Kaau9",
            ec.EllipticCurvePublicKey,
        ),
        (
            "did:key:z82LkvCwHNreneWpsgPEbV3gu1C6NFJEBg4srfJ5gdxEsMGRJUz2sG9FE42shbn2xkZJh54",
            ec.EllipticCurvePublicKey,
        ),
    ],
)
def test_cryptographic_identity_from_did_key(did, key_type):
    identifier = CryptographicIdentifier.from_did_key(did)

    assert identifier.controller == did
    assert isinstance(identifier.public_key, key_type)

    a, b = identifier.as_tuple()
    identifier = CryptographicIdentifier.from_tuple(a, b)

    assert identifier.controller == did
    assert isinstance(identifier.public_key, key_type)


def test_sign_document_fails_for_rsa():
    secret = CryptographicSecret.from_pem("http://localhost/key", private_key)

    with pytest.raises(ValueError):
        secret.sign_document({"@context": [], "https://boo.example": "value"})


@pytest.mark.parametrize(
    ["multibase", "use_rdfc", "cryptosuite"],
    [
        (ed25519_key, False, "eddsa-jcs-2022"),
        (ed25519_key, True, "eddsa-rdfc-2022"),
        (ecp256_key, False, "ecdsa-jcs-2019"),
        (ecp256_key, True, "ecdsa-rdfc-2019"),
    ],
)
def test_sign_document_has_correct_cryptosuite(multibase, use_rdfc, cryptosuite):
    secret = CryptographicSecret.from_multibase("http://localhost/key", multibase)

    signed = secret.sign_document(
        {"@context": [], "https://boo.example": "value"}, use_rdfc=use_rdfc
    )

    assert signed.get("proof").get("cryptosuite") == cryptosuite

    identifier = CryptographicIdentifier("controller", secret.private_key.public_key())

    assert identifier.verify_document(signed) == "controller"



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import pytest
from unittest.mock import AsyncMock
from bovine import BovineClient
from bovine.activitystreams import OrderedCollection
from .collection_helper import CollectionHelper


# See https://codeberg.org/bovine/bovine/issues/69
async def test_multiple_collection_iteration():
    collection_id = "https://bovine.example/collection"
    items = [collection_id + "/two", collection_id + "/three"]
    client = AsyncMock(BovineClient)
    client.proxy.return_value = OrderedCollection(collection_id, items=items).build()

    collection_helper = CollectionHelper(collection_id, client, resolve=False)

    result = [item async for item in collection_helper]

    assert result == items

    result = [item async for item in collection_helper]

    assert result == items


@pytest.mark.skip("requires instance requests")
async def test_collections():
    # remote = "https://metalhead.club/users/mariusor/following"
    # remote = "FIXME"
    remote = "https://mastodon.social/users/the_milkman/outbox"

    async with BovineClient.from_file("bovine_user.toml") as client:
        collection_helper = CollectionHelper(remote, client, resolve=False)

        async for item in collection_helper:
            print(item)



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import bleach


def print_activity(activity):
    if "type" in activity:
        print(f"Type: {activity['type']:10}, Id: {activity.get('id')}")
        print()

    if "object" in activity and isinstance(activity["object"], dict):
        obj = activity["object"]

        print_object(obj, indent="    ")


def print_object(obj, indent=""):
    if "type" in obj:
        print(f"{indent}Type: {obj['type']}")
        print()

    if "name" in obj and obj["name"]:
        print(f"{indent}Name: {obj['name']}")
        print()
    if "summary" in obj and obj["summary"]:
        print(f"{indent}Summary")
        print(indent + bleach.clean(obj["summary"], tags=[], strip=True))
        print()

    if "content" in obj and obj["content"]:
        print(f"{indent}Content")
        print(indent + bleach.clean(obj["content"], tags=[], strip=True))
        print()



# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import json
import sqlite3
import logging

logger = logging.getLogger(__name__)


class JsonLDCache:
    def __init__(self, filename="bovine_jsonld_cache.sqlite"):
        self.filename = filename

        self.connection = sqlite3.connect(filename)

        cursor = self.connection.execute(
            """CREATE TABLE IF NOT EXISTS jsonld (
            url TEXT PRIMARY KEY,
            context TEXT
        )"""
        )

        cursor.execute("SELECT url, context FROM jsonld")
        items = cursor.fetchall()

        self.cache = {x[0]: json.loads(x[1]) for x in items}

        cursor.close()
        self.connection.commit()

    def retrieve(self, url):
        return self.cache.get(url)

    def add(self, url, context):
        self.cache[url] = context

        cursor = self.connection.executemany(
            "INSERT INTO jsonld VALUES (?, ?)", [(url, json.dumps(context))]
        )
        cursor.close()
        self.connection.commit()



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import aiohttp

from dataclasses import dataclass
from bovine.utils import get_gmt_now

from .event_source import EventSource
from .utils import BOVINE_CLIENT_NAME


@dataclass
class BearerAuthClient:
    """Client for using Bearer authentication.

    :param session: The session
    :param bearer_key: The bearer key used in the header `Authorization: Bearer ${bearer_key}`
    """

    session: aiohttp.ClientSession
    bearer_key: str

    async def get(self, url: str, headers: dict = {}):
        """GET of resource.  By default the accept header `application/json` is set. You can override this using `headers`.

        :param url: The target url
        :param headers: additional request headers.
        """
        request_headers = {
            "accept": "application/json",
            "data": get_gmt_now(),
            "user-agent": BOVINE_CLIENT_NAME,
            **headers,
            "authorization": f"Bearer {self.bearer_key}",
        }

        return await self.session.get(url, headers=request_headers)

    async def post(
        self,
        url: str,
        body: str,
        headers: dict = {},
        content_type: str = "application/activity+json",
    ):
        """POST to resource  By default the accept header `application/json` is set. You can override this using `headers`.

        :param url: The target url
        :param body: The request body
        :param headers: additional request headers.
        :param content_type: The content_type of the body
        """
        request_headers = {
            "accept": "application/json",
            "data": get_gmt_now(),
            "user-agent": BOVINE_CLIENT_NAME,
            "content-type": content_type,
            **headers,
            "authorization": f"Bearer {self.bearer_key}",
        }

        return await self.session.post(url, data=body, headers=request_headers)

    def event_source(self, url: str, headers: dict = {}) -> EventSource:
        """Returns an EventSource for the server sent events given by url. Accept header is `text/event-stream` by default

        :param url: The target url
        :param headers: additional request headers.
        """
        request_headers = {
            "accept": "text/event-stream",
            "data": get_gmt_now(),
            "user-agent": BOVINE_CLIENT_NAME,
            **headers,
            "authorization": f"Bearer {self.bearer_key}",
        }
        return EventSource(self.session, url, headers=request_headers)



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import logging

import aiohttp
from typing import Callable, Tuple

from bovine.crypto.http_signature import build_signature
from bovine.crypto.types import CryptographicSecret
from bovine.utils import get_gmt_now

from .event_source import EventSource
from .utils import BOVINE_CLIENT_NAME, host_target_from_url

logger = logging.getLogger(__name__)


async def signed_get(
    session: aiohttp.ClientSession,
    secret: CryptographicSecret,
    url: str,
    headers: dict = {},
) -> aiohttp.ClientResponse:
    logger.debug(f"Signed get with {secret.key_id} on {url}")

    host, target = host_target_from_url(url)
    accept = "application/activity+json"
    date_header = get_gmt_now()

    signature_helper = (
        build_signature(host, "get", target)
        .with_field("date", date_header)
        .with_field("accept", accept)
    )

    signature_header = signature_helper.sign_for_http_draft(secret)

    headers = {
        "user-agent": BOVINE_CLIENT_NAME,
        **headers,
        **signature_helper.headers,
        "signature": signature_header,
    }

    return await session.get(url, headers=headers, allow_redirects=False)


def signed_event_source(
    session: aiohttp.ClientSession,
    secret: CryptographicSecret,
    url: str,
    headers: dict = {},
):
    logger.debug(f"Signed event source with {secret.key_id} on {url}")

    host, target = host_target_from_url(url)
    accept = "text/event-stream"
    date_header = get_gmt_now()

    signature_helper = (
        build_signature(host, "get", target)
        .with_field("date", date_header)
        .with_field("accept", accept)
    )

    signature_header = signature_helper.sign_for_http_draft(secret)

    headers = {
        "user-agent": BOVINE_CLIENT_NAME,
        **headers,
        **signature_helper.headers,
        "signature": signature_header,
    }
    return EventSource(session, url, headers=headers)


async def signed_post(
    session: aiohttp.ClientSession,
    secret: CryptographicSecret,
    url: str,
    body: str,
    digest_method: Callable[[bytes], Tuple[str, str]],
    headers: dict = {},
    content_type=None,
) -> aiohttp.ClientResponse:
    logger.debug(f"Signed post with {secret.key_id} on {url}")

    host, target = host_target_from_url(url)

    if content_type is None:
        content_type = "application/activity+json"
    date_header = get_gmt_now()

    digest_header, digest = digest_method(body.encode("utf-8"))

    signature_helper = (
        build_signature(host, "post", target)
        .with_field("date", date_header)
        .with_field(digest_header, digest)
        .with_field("content-type", content_type)
    )

    signature_header = signature_helper.sign_for_http_draft(secret)

    headers = {
        "user-agent": BOVINE_CLIENT_NAME,
        **headers,
        **signature_helper.headers,
        "signature": signature_header,
    }

    return await session.post(url, data=body, headers=headers)



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import logging
from typing import Optional, Callable, Tuple
from dataclasses import dataclass

import aiohttp

from bovine.crypto import private_key_to_did_key

from . import lookup_did_with_webfinger
from .bearer import BearerAuthClient
from .moo_auth import MooAuthClient
from .signed_http import SignedHttpClient

logger = logging.getLogger(__name__)


@dataclass
class AuthorizationWrapper:
    actor_id: str | None = None
    public_key_url: str | None = None
    access_token: str | None = None
    secret: str | None = None
    domain: str | None = None
    client: BearerAuthClient | MooAuthClient | SignedHttpClient | None = None
    session: aiohttp.ClientSession | None = None
    digest_method: Callable[[bytes], Tuple[str, str]] | None = None

    async def __aenter__(self):
        await self.init()
        return self

    async def __aexit__(self, *args):
        await self.session.close()

    async def init(self, session: Optional[aiohttp.ClientSession] = None):
        self.session = session
        if session is None:
            self.session = aiohttp.ClientSession()

        if self._has_moo_auth():
            await self.with_host_and_ed25519_private_key(
                self.domain, self.secret, session=self.session
            )
        elif self._has_http_signature():
            self.actor_id = self.actor_id
            self.with_http_signature(
                self.public_key_url,
                self.secret,
                session=self.session,
                digest_method=self.digest_method,
            )
        elif self._has_bearer_auth():
            self.actor_id = self.actor_id
            self.with_bearer_token(self.access_token, session=self.session)
        else:
            raise Exception("No known authorization method available")

    async def with_host_and_ed25519_private_key(
        self,
        host: str,
        private_key: str,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        if session is None:
            session = aiohttp.ClientSession()
        did_key = private_key_to_did_key(private_key)

        self.actor_id = await lookup_did_with_webfinger(session, host, did_key)

        if not isinstance(self.actor_id, str):
            logger.error("Failed to lookup actor id")
            raise Exception("Failed to create Moo Auth Client")

        self.client = MooAuthClient(session, did_key, private_key)

        return self

    def with_actor_id(self, actor_id: str):
        self.actor_id = actor_id
        return self

    def with_http_signature(
        self,
        public_key_url: str,
        private_key: str,
        session: Optional[aiohttp.ClientSession] | None = None,
        digest_method: Callable[[bytes], Tuple[str, str]] | None = None,
    ):
        if session is None:
            session = aiohttp.ClientSession()

        self.client = SignedHttpClient(
            session, public_key_url, private_key, digest_method=digest_method
        )

        return self

    def with_bearer_token(
        self,
        access_token: str,
        session: Optional[aiohttp.ClientSession] = None,
    ):
        if session is None:
            session = aiohttp.ClientSession()

        self.client = BearerAuthClient(session, access_token)

        return self

    def _has_http_signature(self):
        return self.actor_id and self.public_key_url and self.secret

    def _has_moo_auth(self):
        return self.domain and self.secret

    def _has_bearer_auth(self):
        return self.actor_id and self.access_token



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import logging
import re

import aiohttp

from bovine.models import JrdData
from .utils import BOVINE_CLIENT_NAME

logger = logging.getLogger(__name__)


async def lookup_with_webfinger(
    session: aiohttp.ClientSession, webfinger_url: str, params: dict
):
    async with session.get(
        webfinger_url, params=params, headers={"user-agent": BOVINE_CLIENT_NAME}
    ) as response:
        if response.status != 200:
            logger.warning(f"{params['resource']} not found using webfinger")
            return None
        text = await response.text("utf-8")
        data = JrdData.model_validate_json(text)

        if data.links is None:
            return None

        for entry in data.links:
            if entry.rel == "self" and re.match(r"application/.*json", entry.type):
                return entry.href

    return None



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from . import Collection


class CollectionHelper:
    """Provides the ability to iterate over a Collection or OrderedCollection

    Usage:

    ```python
    async for item in CollectionHelper(collection_id, bovine_client):
        do_something(item)
    ```

    By setting resolve=True, items are always returned as a dictionary.
    Otherwise, they are returned as a string or dictionary depending on how
    the data is provided by the remote server."""

    def __init__(self, collection_id: str, actor, resolve=False):
        self.actor = actor
        self.collection_id = collection_id
        self.collection = collection_id
        self.items: list = []
        self.resolve = resolve
        self.fetched_first = False

    def update_items(self):
        if not isinstance(self.collection, dict):
            return False
        if "orderedItems" in self.collection:
            self.items = [*self.collection["orderedItems"]]
            return len(self.items) > 0
        if "items" in self.collection:
            self.items = [*self.collection["items"]]
            return len(self.items) > 0
        return False

    async def resolve_item(self, item):
        if not self.resolve or isinstance(item, dict):
            return item
        return await self.actor.proxy(item)

    def __aiter__(self):
        return self

    async def __anext__(self):
        if len(self.items) > 0:
            return await self.resolve_item(self.items.pop(0))
        if isinstance(self.collection, str):
            self.collection = await self.actor.proxy(self.collection)
            if self.update_items():
                return await self.resolve_item(self.items.pop(0))
        if not self.fetched_first and "first" in self.collection:
            self.fetched_first = True
            self.collection = self.collection["first"]
            if self.update_items():
                return await self.resolve_item(self.items.pop(0))
            else:
                return await self.__anext__()
        if "next" in self.collection:
            self.collection = self.collection["next"]
            if self.update_items():
                return await self.resolve_item(self.items.pop(0))
            else:
                return await self.__anext__()

        self.items = []
        self.collection = self.collection_id
        raise StopAsyncIteration

    async def as_collection(self) -> dict:
        """Returns an ActivityStreams collection containing the items represented
        by the collection_id. This is useful for collections spread over multiple
        pages"""

        items = [x async for x in self]
        if isinstance(self.collection, dict):
            collection_id = self.collection.get("part_of")
            if not collection_id:
                collection_id = self.collection["id"]
        else:
            collection_id = self.collection
        return Collection(id=collection_id, items=items).build()



# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import pytest

from .helper import unknown_to_list, reduce_to_ids


@pytest.mark.parametrize(
    ("unknown", "expected"),
    [(None, []), ("test", ["test"]), (["a"], ["a"]), ({"a": "b"}, [{"a": "b"}])],
)
def test_unknown_to_list(unknown, expected):
    assert unknown_to_list(unknown) == expected


@pytest.mark.parametrize(
    ("input_list", "expected"), [([], []), (["a"], ["a"]), ([{"id": "A"}], ["A"])]
)
def test_reduce_to_ids(input_list, expected):
    assert reduce_to_ids(input_list) == expected



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from . import (
    split_into_objects,
    use_context,
    value_from_object,
    with_bovine_context,
)


async def test_json_ld_split_one_item():
    first_id = "https://my_domain/first"
    item = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": first_id,
        "type": "Create",
    }

    subobjects = await split_into_objects(item)

    assert subobjects == [item]


async def test_json_ld_split():
    first_id = "https://my_domain/first"
    second_id = "https://my_domain/second"
    item = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": first_id,
        "type": "Create",
        "object": {
            "type": "Note",
            "id": second_id,
        },
    }

    subobjects = await split_into_objects(item)

    first, second = subobjects
    assert first["id"] == first_id
    assert second["id"] == second_id

    assert first["object"] == second_id


async def test_json_ld_split_subobject():
    first_id = "https://my_domain/first"
    second_id = "https://my_domain/second"
    item = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": first_id,
        "type": "Create",
        "object": {
            "type": "Note",
            "id": second_id,
        },
        "tag": [{"type": "Mention"}],
    }

    subobjects = await split_into_objects(item)

    first, second = subobjects
    assert first["id"] == first_id
    assert second["id"] == second_id

    assert first["object"] == second_id
    assert first["tag"] == {"type": "Mention"}


async def test_json_ld_split_subobject_list():
    first_id = "https://my_domain/first"
    second_id = "https://my_domain/second"
    tags = [{"type": "Mention", "name": "one"}, {"type": "Mention", "name": "two"}]
    item = {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": first_id,
        "type": "Create",
        "object": {
            "type": "Note",
            "id": second_id,
        },
        "tag": tags,
    }

    subobjects = await split_into_objects(item)

    first, second = subobjects
    assert first["id"] == first_id
    assert second["id"] == second_id

    assert first["object"] == second_id
    assert first["tag"] == tags


def test_use_context():
    input_jsonld = {
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            "https://w3id.org/security/v1",
            {
                "manuallyApprovesFollowers": "as:manuallyApprovesFollowers",
                "toot": "http://joinmastodon.org/ns#",
                "featured": {"@id": "toot:featured", "@type": "@id"},
                "featuredTags": {"@id": "toot:featuredTags", "@type": "@id"},
                "alsoKnownAs": {"@id": "as:alsoKnownAs", "@type": "@id"},
                "movedTo": {"@id": "as:movedTo", "@type": "@id"},
                "schema": "http://schema.org#",
                "PropertyValue": "schema:PropertyValue",
                "value": "schema:value",
                "discoverable": "toot:discoverable",
                "Device": "toot:Device",
                "Ed25519Signature": "toot:Ed25519Signature",
                "Ed25519Key": "toot:Ed25519Key",
                "Curve25519Key": "toot:Curve25519Key",
                "EncryptedMessage": "toot:EncryptedMessage",
                "publicKeyBase64": "toot:publicKeyBase64",
                "deviceId": "toot:deviceId",
                "claim": {"@type": "@id", "@id": "toot:claim"},
                "fingerprintKey": {"@type": "@id", "@id": "toot:fingerprintKey"},
                "identityKey": {"@type": "@id", "@id": "toot:identityKey"},
                "devices": {"@type": "@id", "@id": "toot:devices"},
                "messageFranking": "toot:messageFranking",
                "messageType": "toot:messageType",
                "cipherText": "toot:cipherText",
                "suspended": "toot:suspended",
                "Hashtag": "as:Hashtag",
                "focalPoint": {"@container": "@list", "@id": "toot:focalPoint"},
            },
        ],
        "id": "https://somewhere.social/users/someone#main-key",
        "owner": "https://somewhere.social/users/someone",
        "publicKeyPem": """-----BEGIN PUBLIC KEY-----
        XXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXX
        -----END PUBLIC KEY-----
        """,
    }

    output_jsonld = use_context(
        input_jsonld,
        ["https://www.w3.org/ns/activitystreams", "https://w3id.org/security/v1"],
    )

    assert output_jsonld == {
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            "https://w3id.org/security/v1",
        ],
        "id": "https://somewhere.social/users/someone#main-key",
        "owner": "https://somewhere.social/users/someone",
        "publicKeyPem": """-----BEGIN PUBLIC KEY-----
        XXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXX
        -----END PUBLIC KEY-----
        """,
    }


def test_value_from_object():
    value = "my value"
    key = "my_key"

    assert value_from_object({}, key) is None

    assert value == value_from_object({"my_key": value}, key)
    assert value == value_from_object({"my_key": {"@value": value}}, key)


def test_bovine_context():
    data = {
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            "https://w3id.org/security/v1",
        ],
        "id": "https://bovine.social/activitypub/bovine#main-key",
        "owner": "https://bovine.social/activitypub/bovine",
        "publicKeyPem": """-----BEGIN PUBLIC KEY-----
        ...
        -----END PUBLIC KEY-----""",
    }

    transformed = with_bovine_context(data)

    data["@context"] = "about:bovine"

    assert transformed == data



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import logging
import json
import warnings
import http_sf
import hashlib

from urllib.parse import urlparse
from dataclasses import dataclass
from typing import Callable, Awaitable, Tuple
import bovine.utils
from bovine.utils import parse_gmt

from .helper import content_digest_sha256
from .http_signature import HttpSignature
from .signature import parse_signature_header
from .types import CryptographicIdentifier

logger = logging.getLogger(__name__)


@dataclass
class SignatureChecker:
    """Dataclass to encapsulate the logic of checking a HTTP signature

    :param key_retriever: used to resolve the keyId to the cryptographic information"""

    key_retriever: Callable[
        [str], Awaitable[Tuple[str | None, str | None] | CryptographicIdentifier | None]
    ]

    async def validate_signature(
        self,
        method: str,
        url: str,
        headers: dict,
        body: Callable[[], Awaitable[str | bytes]],
    ) -> str | None:
        """Valids a given signature

        :param method: The http method either get or post
        :param url: The url being queried
        :param headers: The request headers
        :param body: A coroutine resolving the the request body. Used for post requests to check the digest.
        """
        if "signature" not in headers:
            logger.debug("Signature not present on request for %s", url)
            logger.debug(json.dumps(dict(headers)))
            return None

        if method.lower() == "post":
            if not self.validate_digest(headers, await body()):
                logger.warning("Validating digest failed")
                return None

        try:
            http_signature = HttpSignature()
            parsed_signature = parse_signature_header(headers["signature"])
            signature_fields = parsed_signature.fields

            if (
                "(request-target)" not in signature_fields
                or "date" not in signature_fields
            ):
                logger.warning("Required field not present in signature")
                return None

            if method.lower() == "post" and all(
                field not in signature_fields for field in ["digest", "content-digest"]
            ):
                logger.warning("Digest not present, but computable")
                return None

            http_date = parse_gmt(headers["date"])
            if not bovine.utils.check_max_offset_now(http_date):
                logger.warning(f"Encountered invalid http date {headers['date']}")
                return None

            for field in signature_fields:
                if field == "(request-target)":
                    method = method.lower()
                    parsed_url = urlparse(url)
                    path = parsed_url.path
                    http_signature.with_field(field, f"{method} {path}")
                else:
                    http_signature.with_field(field, headers[field])

            key_result = await self.key_retriever(parsed_signature.key_id)

            if isinstance(key_result, tuple):
                warnings.warn(
                    "Returning a tuple from key_retriever is deprecated, return a CryptographicIdentifier instead, will be remove in bovine 0.6.0",
                    DeprecationWarning,
                )
                key_result = CryptographicIdentifier.from_pem(*key_result)

            if key_result is None:
                logger.debug(f"Could not retrieve key from {parsed_signature.key_id}")
                return None

            return http_signature.verify_with_identity(
                key_result, parsed_signature.signature
            )

        except Exception as e:
            logger.exception(str(e))
            logger.error(headers)
            return None

        return None

    async def validate_signature_request(self, request) -> str | None:
        """Validates a given signature

        :param request: The request object"""
        return await self.validate_signature(
            request.method, request.url, request.headers, request.get_data
        )

    def validate_digest(self, headers: dict, body: bytes) -> bool:
        """Validates the digest. First checks the `digest` header
        then the `content-digest` header.

        :param headers: The headers of the request
        :param body: The body of the request, currently a warning is raised if body is of type str
        :return: True if digest is present and valid
        """

        if isinstance(body, str):
            warnings.warn("Got body of type str expected bytes")
            body = body.encode("utf-8")

        if "digest" in headers:
            request_digest = headers["digest"]
            request_digest = request_digest[:4].lower() + request_digest[4:]
            digest = content_digest_sha256(body)
            if request_digest != digest:
                logger.warning("Different digest")
                return False

            return True

        if "content-digest" in headers:
            try:
                parsed = http_sf.parse(
                    headers["content-digest"].encode("utf-8"), tltype="dict"
                )
            except Exception as e:
                logger.warning(
                    "Failed to parse header %s with %s",
                    headers["content-digest"],
                    repr(e),
                )
                return False
            valid = False

            if "sha-256" in parsed:
                if parsed["sha-256"][0] == hashlib.sha256(body).digest():
                    valid = True
                else:
                    return False

            if "sha-512" in parsed:
                if parsed["sha-512"][0] == hashlib.sha512(body).digest():
                    valid = True
                else:
                    return False

            return valid

        return False



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

from .controller import Controller
from .multikey import Multikey
from . import Actor


def test_empty_controller():
    controller = Controller()

    assert controller.build() == {}


def test_with_multikey():
    multikey = Multikey(
        id="https://server.example/users/alice#ed25519-key",
        controller="https://server.example/users/alice",
        multibase="z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2",
    )

    controller = Controller(assertion_method=[multikey])

    result = controller.build()

    assert result["@context"] == [
        "https://www.w3.org/ns/did/v1",
        "https://w3id.org/security/multikey/v1",
    ]

    assert len(result["assertionMethod"]) == 1

    key = result["assertionMethod"][0]
    assert (
        key["publicKeyMultibase"] == "z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2"
    )
    assert "@context" not in key


def test_actor_as_controller():
    multikey = Multikey(
        id="https://server.example/users/alice#ed25519-key",
        controller="https://server.example/users/alice",
        multibase="z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2",
    )

    controller = Actor(id="https://actor.example", assertion_method=[multikey])

    result = controller.build()

    assert result["@context"] == [
        "https://www.w3.org/ns/activitystreams",
        "https://www.w3.org/ns/did/v1",
        "https://w3id.org/security/multikey/v1",
    ]

    assert len(result["assertionMethod"]) == 1

    key = result["assertionMethod"][0]
    assert (
        key["publicKeyMultibase"] == "z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2"
    )
    assert "@context" not in key



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from typing import List

from bovine.jsonld import use_context

from .multikey import Multikey


@dataclass
class Controller:
    """Experimental class to represent a controller document
    see [FEP-521a](https://codeberg.org/fediverse/fep/src/branch/main/fep/521a/fep-521a.md)
    """

    assertion_method: List[Multikey] = field(default_factory=list)
    authentication: List[Multikey] = field(default_factory=list)

    def build(self):
        if len(self.assertion_method) == 0:
            return {}

        result = {
            "@context": [
                "https://www.w3.org/ns/did/v1",
            ],
            "assertionMethod": [key.build() for key in self.assertion_method],
        }

        return use_context(
            result,
            [
                "https://www.w3.org/ns/did/v1",
                "https://w3id.org/security/multikey/v1",
            ],
        )



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from unittest.mock import AsyncMock

import aiohttp

from bovine.crypto.test import private_key
from bovine.crypto.helper import content_digest_sha256_rfc_9530

from .signed_http import SignedHttpClient


async def test_activity_pub_client_get():
    session = AsyncMock(aiohttp.ClientSession)
    url = "https://test_domain/test_path"
    public_key_url = "public_key_url"
    session = AsyncMock(aiohttp.ClientSession)
    session.get = AsyncMock()

    client = SignedHttpClient(session, public_key_url, private_key)

    await client.get(url)

    session.get.assert_awaited_once()


async def test_activity_pub_client_post():
    session = AsyncMock(aiohttp.ClientSession)
    url = "https://test_domain/test_path"
    public_key_url = "public_key_url"
    session = AsyncMock(aiohttp.ClientSession)
    session.post = AsyncMock()

    client = SignedHttpClient(session, public_key_url, private_key)

    await client.post(url, "hello")

    session.post.assert_awaited_once()

    args = session.post.await_args_list[0]
    headers = args[1]["headers"]
    assert headers["digest"] == "sha-256=LPJNul+wow4m6DsqxbninhsWHlwfp0JecwQzYpOLmCQ="


async def test_activity_pub_client_post_rfc_9530():
    session = AsyncMock(aiohttp.ClientSession)
    url = "https://test_domain/test_path"
    public_key_url = "public_key_url"
    session = AsyncMock(aiohttp.ClientSession)
    session.post = AsyncMock()

    client = SignedHttpClient(
        session,
        public_key_url,
        private_key,
        digest_method=content_digest_sha256_rfc_9530,
    )

    await client.post(url, '{"cows": "good"}')

    session.post.assert_awaited_once()

    args = session.post.await_args_list[0]
    headers = args[1]["headers"]
    assert (
        headers["content-digest"]
        == "sha-256=:MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=:"
    )
    assert headers["content-type"] == "application/activity+json"



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives import serialization

from bovine.models import DataIntegrityProof

from .helper import (
    content_digest_sha256,
    did_key_to_public_key,
    verify_signature,
    multibase_decode,
    split_according_to_fep8b32,
    integrity_proof_to_pure_proof_dict,
    content_digest_sha256_rfc_9530,
)
from .test import private_key, public_key
from .types import CryptographicSecret


def test_content_digest_sha256():
    digest = content_digest_sha256("content")

    assert digest == "sha-256=7XACtDnprIRfIjV9giusFERzD722AW0+yUMil7nsn3M="


def test_did_to_public_key():
    did_example = "did:key:z6MkiTBz1ymuepAQ4HEHYSF1H8quG5GLVVQR3djdX3mDooWp"

    public_key = did_key_to_public_key(did_example)

    assert isinstance(public_key, ed25519.Ed25519PublicKey)
    assert (
        public_key.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
        )
        == b";j'\xbc\xce\xb6\xa4-b\xa3\xa8\xd0*o\rse2\x15w\x1d\xe2C\xa6:\xc0H\xa1\x8bY\xda)"
    )


def test_crypto_sign_verify():
    message = "secret"

    secret = CryptographicSecret.from_pem("key", private_key)
    signature = secret.sign(message)

    assert verify_signature(public_key, message, signature)


def test_crypto_sign_verify_failure():
    message = "secret"

    assert not verify_signature(public_key, message, "")


def test_split_according_to_fep8b32():
    doc_base = {
        "@context": [
            "https://www.w3.org/ns/activitystreams",
            "https://w3id.org/security/data-integrity/v1",
        ],
        "type": "Create",
        "actor": "https://server.example/users/alice",
        "object": {"type": "Note", "content": "Hello world"},
    }
    pure_proof = {
        "type": "DataIntegrityProof",
        "cryptosuite": "eddsa-jcs-2022",
        "verificationMethod": "https://server.example/users/alice#ed25519-key",
        "proofPurpose": "assertionMethod",
        "created": "2023-02-24T23:36:38Z",
    }
    signature = "z3sXaxjKs4M3BRicwWA9peyNPJvJqxtGsDmpt1jjoHCjgeUf71TRFz56osPSfDErszyLp5Ks1EhYSgpDaNM977Rg2"

    input_doc = {**doc_base, "proof": {**pure_proof, "proofValue": signature}}

    doc, proof, sig = split_according_to_fep8b32(input_doc)

    assert doc == doc_base
    assert integrity_proof_to_pure_proof_dict(proof) == pure_proof
    assert sig == multibase_decode(signature)


def test_split_according_to_fep8b32_context_in_credential():
    input_doc = {
        "@context": ["https://www.w3.org/2018/credentials/v1"],
        "id": "urn:uuid:09327695-56d3-4ef2-890c-d63b2b5e707f",
        "type": ["VerifiableCredential"],
        "credentialSubject": {
            "id": "did:key:z6MktKwz7Ge1Yxzr4JHavN33wiwa8y81QdcMRLXQsrH9T53b"
        },
        "issuer": "did:key:z6MkgND5U5Kedizov5nxeh2ZCVUTDRSmAfbNqPhzCq8b72Ra",
        "issuanceDate": "2020-03-16T22:37:26.544Z",
        "proof": {
            "@context": ["https://w3id.org/security/data-integrity/v1"],
            "type": "DataIntegrityProof",
            "proofPurpose": "assertionMethod",
            "proofValue": "z4ncyz44KMkma4Z4h3iVRx6W2tB7xVbMNGk8gc9BRtqNWWsixNazdshm1aZADEH3vNJxd9RMKBTMqhSCcSEH6tn8V",
            "verificationMethod": "did:key:z6MkgND5U5Kedizov5nxeh2ZCVUTDRSmAfbNqPhzCq8b72Ra#z6MkgND5U5Kedizov5nxeh2ZCVUTDRSmAfbNqPhzCq8b72Ra",
            "created": "2024-02-05T10:21:42.315Z",
            "cryptosuite": "eddsa-2022",
        },
    }

    doc, proof, sig = split_according_to_fep8b32(input_doc)

    assert isinstance(proof, DataIntegrityProof)
    assert proof.field_context == ["https://w3id.org/security/data-integrity/v1"]


def test_content_digest_sha256_rfc_9530():
    header, result = content_digest_sha256_rfc_9530(b'{"hello": "world"}\n')

    assert header == "content-digest"
    assert result == "sha-256=:RK/0qy18MlBSVnWgjwz6lZEWjP/lF5HF9bvEF8FabDg=:"



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import base64
import logging

from dataclasses import dataclass
from cryptography.exceptions import InvalidSignature

from cryptography.hazmat.primitives.asymmetric import ed25519, rsa, padding, ec, utils
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import (
    load_pem_public_key,
    load_pem_private_key,
)
from typing import Union, Tuple

import bovine.jsonld
import bovine.utils
from bovine.utils import pydantic_to_json
from bovine.models import DataIntegrityProof, Multikey
from .helper import (
    split_according_to_fep8b32,
    jcs_sha256,
    jcs_sha384,
    rdfc_sha256,
    rdfc_sha384,
    integrity_proof_to_pure_proof_dict,
    ensure_data_integrity_context,
    curve_to_hashing_algorithm_and_half_key_length,
)

from .multibase import (
    multibase_58btc_encode,
    decode_multibase_public_key,
    encode_public_key_to_multibase,
    multibase_to_private_key,
)


logger = logging.getLogger(__name__)


@dataclass
class CryptographicSecret:
    """Represents a cryptographic secret. Such a secret is composed
    of the private key and the URI that resolves to the material, one
    can construct the appropriate cryptographic identifier from.

    :param key_id: The URI, where the corresponding public key and controller
        can be retrieved
    :param private_key: The signing material
    """

    key_id: str
    private_key: Union[ed25519.Ed25519PrivateKey, rsa.RSAPrivateKey]

    def sign(self, message: str | bytes):
        """Signs the message.

        Currently only implemented for RSA: Uses PKCS1v15 padding and SHA256
        hashes. Returns the signature as base64 encoded.

        **Warning**: Interface might change, to enable specifying encoding
        of the signature.

        :param message: The message to sign as UTF-8 encoded string."""

        if isinstance(message, str):
            message = message.encode("utf-8")

        if isinstance(self.private_key, rsa.RSAPrivateKey):
            return base64.standard_b64encode(
                self.private_key.sign(
                    message,
                    padding.PKCS1v15(),
                    hashes.SHA256(),
                )
            ).decode("utf-8")
        if isinstance(self.private_key, ed25519.Ed25519PrivateKey):
            return multibase_58btc_encode(self.private_key.sign(message))
        if isinstance(self.private_key, ec.EllipticCurvePrivateKey):
            algorithm, length = curve_to_hashing_algorithm_and_half_key_length(
                self.private_key.curve
            )
            signature_dss = self.private_key.sign(message, algorithm)
            r, s = utils.decode_dss_signature(signature_dss)
            signature = r.to_bytes(length) + s.to_bytes(length)
            return multibase_58btc_encode(signature)

        raise ValueError("Unknown key type in private key")

    def sign_document(
        self,
        document: dict,
        use_rdfc: bool = False,
        proof_purpose="assertionMethod",
    ) -> dict:
        """Signs the current document according to the procedure
        outlined in [FEP-8b32](https://codeberg.org/fediverse/fep/src/branch/main/fep/8b32/fep-8b32.md). We support signing with
        [eddsa-jcs-2022 / eddsa-rdfc-2022](https://www.w3.org/TR/vc-di-eddsa/)
        and [ecdsa-jcs-2019 / eddsa-edfc-2019](https://www.w3.org/TR/vc-di-ecdsa/).
        The cryptosuite is chosen depending on the type of signing material
        and the `use_rdfc` parameter

        :param document: The document to sign
        :param use_rdfc: Set to `True` to use `eddsa-rdfc-2022` or `ecdsa-rdfc-2019`
        :param proof_purpose: Purpose of the proof
        """

        match [self.private_key, use_rdfc]:
            case [ed25519.Ed25519PrivateKey(), True]:
                cryptosuite = "eddsa-rdfc-2022"
                to_hash = rdfc_sha256
            case [ed25519.Ed25519PrivateKey(), False]:
                cryptosuite = "eddsa-jcs-2022"
                to_hash = jcs_sha256
            case [ec.EllipticCurvePrivateKey(), True]:
                cryptosuite = "ecdsa-rdfc-2019"
                if isinstance(self.private_key.curve, ec.SECP256R1):
                    to_hash = rdfc_sha256
                else:
                    to_hash = rdfc_sha384
            case [ec.EllipticCurvePrivateKey(), False]:
                cryptosuite = "ecdsa-jcs-2019"
                if isinstance(self.private_key.curve, ec.SECP256R1):
                    to_hash = jcs_sha256
                else:
                    to_hash = jcs_sha384
            case _:
                raise ValueError("Unsupported Key Format for signing documents")

        document_to_sign = ensure_data_integrity_context(document)

        created = bovine.utils.now_isoformat()
        proof = DataIntegrityProof(
            created=created,
            cryptosuite=cryptosuite,
            proofPurpose=proof_purpose,
            type="DataIntegrityProof",
            verificationMethod=self.key_id,
        )
        digest = to_hash(
            pydantic_to_json(proof),
            context=document_to_sign["@context"],
        ) + to_hash(document_to_sign, context=document_to_sign["@context"])
        proof.proofValue = self.sign(digest)

        return {
            **document_to_sign,
            "proof": pydantic_to_json(proof),
        }

    @classmethod
    def from_pem(cls, key_id: str, pem: str):
        """Creates a CryptographicSecret from a PEM encoded private key"""
        return cls(key_id, load_pem_private_key(pem.encode("utf-8"), password=None))

    @classmethod
    def from_multibase(cls, key_id: str, multibase: str):
        """Creates a CryptographicSecret from multibase encoded
        Ed25519 private key and key_id"""
        return cls(key_id, multibase_to_private_key(multibase))


@dataclass
class CryptographicIdentifier:
    """Represents a cryptographic identifier. The usage is: If an object is
    signed by `public_key`, then it is authored by `controller`. In order
    to discover which `CryptographicIdentifier` to use, one resolves another
    identifier `key_id`, which yields either a Multikey or a publicKey object, which
    can then be resolved into a CryptographicIdentifier.

    One should never need to directly access the properties of this class, instead
    verify returns the controller, if and only if the signature is valid.

    :param controller: The URI of the actor that controls the public key
    :param public_key: Public key used to verify signatures
    """

    controller: str
    public_key: Union[
        ed25519.Ed25519PublicKey, rsa.RSAPublicKey, ec.EllipticCurvePublicKey
    ]

    def verify(self, message: str | bytes, signature: str | bytes) -> str | None:
        """Verifies that `signature` is a correct signature for the given message.

        **Warning**: Interface might change, to enable specifying encoding
        of the signature.

        :param message: The message string.
        :param signature: The signature

        :return: If the signature is valid the corresponding controller,
            otherwise null.
        """
        # Doing the encoding here is probably wrong ...
        # All these things are awkward

        if isinstance(signature, str):
            signature = base64.standard_b64decode(signature)

        if isinstance(message, str):
            message = message.encode("utf-8")

        if isinstance(self.public_key, rsa.RSAPublicKey):
            try:
                self.public_key.verify(
                    signature,
                    message,
                    padding.PKCS1v15(),
                    hashes.SHA256(),
                )
                return self.controller
            except InvalidSignature:
                return None

        if isinstance(self.public_key, ed25519.Ed25519PublicKey):
            try:
                self.public_key.verify(signature, message)
                return self.controller
            except InvalidSignature:
                return None

        if isinstance(self.public_key, ec.EllipticCurvePublicKey):
            try:
                algorithm, length = curve_to_hashing_algorithm_and_half_key_length(
                    self.public_key.curve
                )

                r = int.from_bytes(signature[:length])
                s = int.from_bytes(signature[length:])

                logger.debug("R %d", r)
                logger.debug("S %d", s)

                signature_dss = utils.encode_dss_signature(r, s)

                logger.debug("Signature DSS length %d", len(signature_dss))

                self.public_key.verify(signature_dss, message, algorithm)
                return self.controller
            except InvalidSignature:
                return None

        raise ValueError("Unknown key type in public_key")

    def verify_document(self, document: dict):
        """Verifies that document has a valid signature according to FEP-8b32.
        We note that in order to verify a document signed using FEP-8b32, one
        will already need to parse it sufficiently to extract the controller,
        so the CryptographicIdentifier can be created.

        __Beware__: Verification with P-386 keys might be broken.

        :param document: The document to verify"""
        pure_doc, proof_doc, signature = split_according_to_fep8b32(document)
        pure_proof = integrity_proof_to_pure_proof_dict(proof_doc)

        match proof_doc.cryptosuite:
            case "eddsa-jcs-2022" | "ecdsa-jcs-2019":
                to_hash = jcs_sha256
                if isinstance(
                    self.public_key, ec.EllipticCurvePublicKey
                ) and isinstance(self.public_key.curve, ec.SECP384R1):
                    to_hash = jcs_sha384
            case "eddsa-rdfc-2022" | "ecdsa-rdfc-2019":
                to_hash = rdfc_sha256
                if isinstance(
                    self.public_key, ec.EllipticCurvePublicKey
                ) and isinstance(self.public_key.curve, ec.SECP384R1):
                    to_hash = rdfc_sha384
                    logger.info("USING rdfc_sha384")
            case _:
                logger.warning("Got the unknown cryptosuite %s", proof_doc.cryptosuite)
                raise ValueError(f"Unknown cryptosuite '{proof_doc.cryptosuite}'")

        digest_one = to_hash(pure_proof, context=pure_doc["@context"])
        digest_two = to_hash(pure_doc, context=pure_doc["@context"])
        logger.debug("Digest Proof %s", digest_one.hex())
        logger.debug("Digest Document %s %d", digest_two.hex(), len(digest_two))
        digest = digest_one + digest_two

        return self.verify(digest, signature)

    def as_tuple(self) -> Tuple[str, str]:
        """Transforms the CryptographicIdentifier into a tuple

        :return: controller, multibase/multicodec encoded public key"""
        public_key = encode_public_key_to_multibase(self.public_key)

        return (self.controller, public_key)

    @classmethod
    def from_pem(cls, public_key: str, owner: str):
        """Creates a CryptographicIdentifier from a pem encoded public key and the controller

        ```pycon
        >>> public_key_pem = '-----BEGIN PUBLIC KEY-----\\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA15vhFdK272bbGDzLtypo\\n4Nn8mNFY3YSLnrAOX4zZKkNmWmgypgDP8qXjNiVsBf8f+Yk3tHDs58LMf8QDSP09\\nA+zrlWBHN1rLELn0JBgqT9xj8WSobDIjOjFBAy4FKUko7k/IsYwTl/Vnx1tykhPR\\n1UzbaNqN1yQSy0zGbIce/Xhqlzm6u+twyuHVCtbGPcPh7for5o0avKdMwhAXpWMr\\nNoc9L2L/\\n9h3UgoePgAvCE6HTPXEBPesUBlTULcRxMXIZJ7P6eMkb2pGUCDlVF4EN\\nvcxZAG8Pb7HQp9nzVwK4OXZclKsH1YK0G8oBGTxnroBtq7cJbrJvqNMNOO5Yg3cu\\n6QIDAQAB\\n-----END PUBLIC KEY-----'
        >>> identifier = CryptographicIdentifier.from_pem(public_key_pem, 'https://com.example/issuer/123')
        >>> identifier.controller
        'https://com.example/issuer/123'

        ```

        """
        if public_key is None:
            return None
        return cls(
            controller=owner, public_key=load_pem_public_key(public_key.encode("utf-8"))
        )

    @classmethod
    def from_multikey(cls, multikey: dict):
        """Creates a CryptographicIdentifier from a Multikey, see

        * [FEP-521a: Representing actor's public keys](https://codeberg.org/fediverse/fep/src/branch/main/fep/521a/fep-521a.md)

        Example:

        ```pycon
        >>> identifier = CryptographicIdentifier.from_multikey({
        ...     "id": "https://server.example/users/alice#ed25519-key",
        ...     "type": "Multikey",
        ...     "controller": "https://server.example/users/alice",
        ...     "publicKeyMultibase": "z6MkrJVnaZkeFzdQyMZu1cgjg7k1pZZ6pvBQ7XJPt4swbTQ2"
        ... })
        >>> identifier.controller
        'https://server.example/users/alice'

        ```
        """
        parsed = Multikey.model_validate(multikey)
        return cls.from_tuple(parsed.controller, parsed.publicKeyMultibase)

    @classmethod
    def from_tuple(cls, controller: str, multibase_public_key: str):
        """Creates a CryptographicIdentifier from a tuple

        :param controller: The controller URI
        :param multibase_public_key: The public key encoded using multibase/multicodex
        """
        public_key = decode_multibase_public_key(multibase_public_key)
        if public_key:
            return cls(
                controller=controller,
                public_key=public_key,
            )
        raise ValueError("Unsupported public key format")

    @classmethod
    def from_did_key(cls, did_key: str):
        """Creates a cryptographic identifier from a did:key
        The controller is then the did:key and the public key
        the encoded public key. See [The did:key Method](https://w3c-ccg.github.io/did-method-key/) for details. Currently supported: Ed25519, RSA, P-256, and P-384 keys.

        ```pycon
        >>> identifier = CryptographicIdentifier.from_did_key("did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5")
        >>> identifier.controller
        'did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5'

        ```

        :param did_key: The did key
        """

        if did_key.startswith("did:key:"):
            return cls.from_tuple(did_key, did_key[8:])

        raise ValueError("Invalid did key format")

    @classmethod
    def from_public_key(cls, data: dict):
        """Creates a Cryptographic identifier from a publicKey object, example:

        ```pycon
        >>> public_key = {
        ...    "id": "https://com.example/issuer/123#main-key",
        ...    "owner": "https://com.example/issuer/123",
        ...    "publicKeyPem": '-----BEGIN PUBLIC KEY-----\\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA15vhFdK272bbGDzLtypo\\n4Nn8mNFY3YSLnrAOX4zZKkNmWmgypgDP8qXjNiVsBf8f+Yk3tHDs58LMf8QDSP09\\nA+zrlWBHN1rLELn0JBgqT9xj8WSobDIjOjFBAy4FKUko7k/IsYwTl/Vnx1tykhPR\\n1UzbaNqN1yQSy0zGbIce/Xhqlzm6u+twyuHVCtbGPcPh7for5o0avKdMwhAXpWMr\\nNoc9L2L/\\n9h3UgoePgAvCE6HTPXEBPesUBlTULcRxMXIZJ7P6eMkb2pGUCDlVF4EN\\nvcxZAG8Pb7HQp9nzVwK4OXZclKsH1YK0G8oBGTxnroBtq7cJbrJvqNMNOO5Yg3cu\\n6QIDAQAB\\n-----END PUBLIC KEY-----'
        ... }
        >>> identifier = CryptographicIdentifier.from_public_key(public_key)
        >>> identifier.controller
        'https://com.example/issuer/123'

        ```

        """
        controller = data.get("owner")
        public_key = data.get("publicKeyPem")

        if isinstance(public_key, dict):
            public_key = public_key.get("@value")

        if controller is None or public_key is None:
            raise ValueError(
                "Expected parameters owner and publicKeyPem to be present when parsing publicKey"
            )

        return cls(
            controller=controller,
            public_key=load_pem_public_key(public_key.encode("utf-8")),
        )



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from . import activity_pub_object_id_from_html_body, parse_fediverse_handle


def test_parse_fediverse_handle():
    assert parse_fediverse_handle("account") == ("account", None)
    assert parse_fediverse_handle("account@domain") == ("account", "domain")
    assert parse_fediverse_handle("acct:account@domain") == ("account", "domain")

    assert parse_fediverse_handle("account@domain@@@") == ("account", "domain@@@")
    assert parse_fediverse_handle("@account@domain@@@") == ("account", "domain@@@")
    assert parse_fediverse_handle("@account") == ("account", None)


def test_activity_pub_object_id_from_html_body():
    object_id = "https://bovine.example/object_id"
    body = f"""\
<!DOCTYPE HTML>
<html>
<head>
<link href="{object_id}"
rel="alternate"
type="application/activity+json" />
</head>
<body>
Test case 1; link header in head
</body>
</html>
"""

    result = activity_pub_object_id_from_html_body(body)

    assert result == object_id



# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from typing import List


def unknown_to_list(obj) -> list:
    match obj:
        case dict() | str():
            return [obj]
        case list():
            return obj
    return []


def reduce_to_ids(list_obj: List) -> List[str]:
    return [x if isinstance(x, str) else x.get("id") for x in list_obj]



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import asyncio
import json
import click

from asyncstdlib.itertools import islice
from ptpython.repl import embed

from bovine import BovineClient
from bovine.activitystreams.utils.print import print_activity


def config(repl):
    repl.use_code_colorscheme("dracula")
    repl.enable_output_formatting = True


async def repl(client):
    async with client:
        activity_factory, object_factory = client.factories
        print("The variable client contains your BovineClient")
        print("The variables activity_factory and object_factory")
        print("contain the corresponding objects")
        print("With await client.inbox() and await client.outbox()")
        print("one can interface with these two")
        print()
        await embed(
            globals=globals(),
            locals=locals(),
            return_asyncio_coroutine=True,
            patch_stdout=True,
            configure=config,
        )


async def show_inbox(client: BovineClient, max_number: int = 10, summary: bool = False):
    async with client:
        await display_box(await client.inbox(), max_number, summary)


async def show_outbox(
    client: BovineClient, max_number: int = 10, summary: bool = False
):
    async with client:
        await display_box(await client.outbox(), max_number, summary)


async def display_box(box, max_number: int, summary: bool):
    async for item in islice(box, max_number):
        if summary:
            print_activity(item)
        else:
            print(json.dumps(item, indent=2))
            print()


@click.command(help="Opens a REPL with preloaded BovineClient client")
@click.option("--domain", help="Domain the actor can be found on")
@click.option("--secret", help="Secret associated with the account")
@click.option("--config_file", help="Toml fail containing domain and secret")
@click.option("--inbox", is_flag=True, default=False, help="Display the inbox")
@click.option("--outbox", is_flag=True, default=False, help="Display the outbox")
@click.option(
    "--summary",
    is_flag=True,
    default=False,
    help="Display a summary of the activity instead of the full json",
)
@click.option(
    "--max_number",
    type=int,
    help="Number of elements to display in inbox or outbox",
    default=20,
)
def main(domain, secret, config_file, inbox, outbox, summary, max_number):
    if config_file:
        client = BovineClient.from_file(config_file)
    elif domain and secret:
        client = BovineClient(domain=domain, secret=secret)
    else:
        default_config_file = "bovine_user.toml"
        client = BovineClient.from_file(default_config_file)

        print(f"Config file not specified using fallback value '{default_config_file}'")

    if inbox:
        asyncio.run(show_inbox(client, max_number=max_number, summary=summary))
    elif outbox:
        asyncio.run(show_outbox(client, max_number=max_number, summary=summary))
    else:
        loop = asyncio.new_event_loop()
        loop.run_until_complete(repl(client))


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import asyncio
import click

from .utils.msg import process, validate


@click.command()  # help="Send a message through your ActivityPub Actor")
@click.option(
    "--secret",
    help="Secret corresponding to a did-key deposited with your actor",
)
@click.option("--host", help="Hostname your actor is on")
@click.option(
    "--to",
    multiple=True,
    help="Recipients either as FediVerse handle or actor id",
    default=[],
)
@click.option("--public", is_flag=True, default=False)
@click.argument("message", nargs=-1)
def main(secret, host, to, public, message):
    validate(to, public)
    asyncio.run(process(secret, host, to, public, message))


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import aiohttp
from dataclasses import dataclass

from bovine.crypto.helper import content_digest_sha256
from bovine.crypto.http_signature import build_signature
from bovine.utils import get_gmt_now

from .event_source import EventSource
from .utils import BOVINE_CLIENT_NAME, host_target_from_url


@dataclass
class MooAuthClient:
    """Client for using Moo-Auth-1 authentication

    :param session: The session
    :param did_key: The did key, i.e. `did:key:z...`
    :param private_key: private key corresponding to did_key
    """

    session: aiohttp.ClientSession
    did_key: str
    private_key: str

    async def get(self, url: str, headers: dict = {}):
        """GET for resource

        :param url: url to query
        :param headers: Additional headers"""
        host, target = host_target_from_url(url)

        accept = "application/activity+json"
        date_header = get_gmt_now()

        signature_helper = build_signature(host, "get", target).with_field(
            "date", date_header
        )
        signature_header = signature_helper.ed25519_sign(self.private_key)

        headers = {
            "accept": accept,
            "user-agent": BOVINE_CLIENT_NAME,
            **headers,
            **signature_helper.headers,
            "authorization": f"Moo-Auth-1 {self.did_key}",
            "x-moo-signature": signature_header,
        }

        return await self.session.get(url, headers=headers)

    async def post(
        self,
        url: str,
        body: str,
        headers: dict = {},
        content_type: str = "application/activity+json",
    ):
        """POST to resource

        :param url: The target url
        :param body: The request body
        :param headers: additional request headers.
        :param content_type: The content_type of the body
        """
        host, target = host_target_from_url(url)
        accept = "application/activity+json"
        date_header = get_gmt_now()

        digest = content_digest_sha256(body)

        signature_helper = (
            build_signature(host, "post", target)
            .with_field("date", date_header)
            .with_field("digest", digest)
        )
        signature_header = signature_helper.ed25519_sign(self.private_key)

        headers = {
            "accept": accept,
            "content-type": content_type,
            "user-agent": BOVINE_CLIENT_NAME,
            **headers,
            **signature_helper.headers,
            "authorization": f"Moo-Auth-1 {self.did_key}",
            "x-moo-signature": signature_header,
        }

        return await self.session.post(url, data=body, headers=headers)

    def event_source(self, url: str, headers: dict = {}):
        """Returns an event source

        :param url: url of event source
        :param headers: additional headers"""
        host, target = host_target_from_url(url)
        date_header = get_gmt_now()
        accept = "text/event-stream"
        signature_helper = build_signature(host, "get", target).with_field(
            "date", date_header
        )
        signature_header = signature_helper.ed25519_sign(self.private_key)

        headers = {
            "accept": accept,
            "user-agent": BOVINE_CLIENT_NAME,
            **headers,
            **signature_helper.headers,
            "authorization": f"Moo-Auth-1 {self.did_key}",
            "x-moo-signature": signature_header,
        }

        return EventSource(self.session, url, headers=headers)



# SPDX-FileCopyrightText: 2023 Helge
#
# SPDX-License-Identifier: MIT

# generated by datamodel-codegen:
#   filename:  models.schema.json
#   timestamp: 2024-02-05T11:00:10+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import AnyUrl, AwareDatetime, BaseModel, ConfigDict, Field, RootModel


class Model(RootModel[Any]):
    root: Any


class DataIntegrityProof(BaseModel):
    """
    A data integrity proof provides information about the proof mechanism, parameters required to verify that proof, and the proof value itself. See https://www.w3.org/TR/vc-data-integrity/#proofs
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    field_context: Optional[Any] = Field(None, alias="@context")
    """
    Optional `@context` parameter
    """
    id: Optional[str] = None
    """
    An optional identifier for the proof, which MUST be a URL, such as a UUID as a URN (urn:uuid:6a1676b8-b51f-11ed-937b-d76685a20ff5).
    """
    type: str
    """
    The specific proof type used for the cryptographic proof MUST be specified as a string that maps to a URL. Examples of proof types include DataIntegrityProof and Ed25519Signature2020. Proof types determine what other fields are required to secure and verify the proof. 
    """
    cryptosuite: str
    """
    fixme
    """
    created: Optional[AwareDatetime] = None
    """
    The date and time the proof was created 
    """
    expires: Optional[AwareDatetime] = None
    """
    dateTimeStamp string specifying when the proof expires. 
    """
    verificationMethod: AnyUrl
    """
    The means and information needed to verify the proof MUST be specified as a string that maps to a URL. An example of a verification method is a link to a public key which includes cryptographic material that is used by a verifier during the verification process. 
    """
    proofPurpose: str
    """
    The proof purpose acts as a safeguard to prevent the proof from being misused by being applied to a purpose other than the one that was intended. 
    """
    proofValue: Optional[str] = None
    """
    A string value that contains the base-encoded binary data necessary to verify the digital proof using the verificationMethod specified. 
    """
    nonce: Optional[str] = None
    domain: Optional[Union[str, List[str]]] = None
    challenge: Optional[str] = None
    previousProof: Optional[Union[str, List[str]]] = None


class Type(Enum):
    """
    The type, must be Multikey
    """

    Multikey = "Multikey"


class Multikey(BaseModel):
    """
    Representation of a cryptographic identifier as specified by Verifiable Credentials Data Integrity [here](https://www.w3.org/TR/vc-data-integrity/#multikey)
    """

    model_config = ConfigDict(
        extra="forbid",
    )
    field_context: Optional[Any] = Field(None, alias="@context")
    id: Optional[str] = None
    """
    optional id
    """
    type: Type
    """
    The type, must be Multikey
    """
    controller: str
    """
    The controller, i.e. the person holding the corresponding secret
    """
    publicKeyMultibase: str
    """
    The public key encoded in base 58 with an appropriate prefix
    """
    expires: Optional[AwareDatetime] = None
    """
    dateTimeStamp string specifying when the key expires. 
    """
    revoked: Optional[AwareDatetime] = None
    """
    dateTimeStamp string specifying when the key was revoked. 
    """


class JrdLink(BaseModel):
    """
    See [RFC 7033](https://www.rfc-editor.org/rfc/rfc7033.html).  The XRD Link
    """

    model_config = ConfigDict(
        extra="allow",
    )
    rel: Optional[str] = None
    """
    rel
    """
    type: Optional[str] = None
    """
    type
    """
    href: Optional[str] = None
    """
    when used with the 'href' attribute, conveys a link relation between the host described by the document and a common target URI
    """
    titles: Optional[Dict[str, Any]] = None
    """
    titles
    """
    properties: Optional[Dict[str, Any]] = None
    """
    properties
    """
    template: Optional[str] = None
    """
    template attribute conveys a relation whose context is an individual resource within the host-meta document scope,
    """


class JrdData(BaseModel):
    """
    See [RFC 6415](https://www.packetizer.com/rfc/rfc6415/)
    """

    model_config = ConfigDict(
        extra="allow",
    )
    subject: Optional[AnyUrl] = None
    """
    The subject
    """
    expires: Optional[AwareDatetime] = None
    """
    expiration date time
    """
    aliases: Optional[List[AnyUrl]] = None
    """
    value a string array containing the values of each element in order
    """
    properties: Optional[Dict[str, Any]] = None
    """
    value an object with each element included as a name/value pair with the value of the type attribute as name, and element value included as a string value.
    """
    links: Optional[List[JrdLink]] = None
    """
     a single name/value pair with the name 'links', and value an array with each element included as an object
    """



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from . import BovineClient


def test_bovine_client_host():
    actor = BovineClient().with_actor_id("https://domain.tld/users/someone")

    assert actor.host == "domain.tld"



# SPDX-FileCopyrightText: 2024 Helge
#
# SPDX-License-Identifier: MIT

import pytest
from unittest.mock import AsyncMock

from .signature_checker import SignatureChecker


@pytest.mark.parametrize(
    "headers",
    [
        {"digest": "sha-256=MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k="},
        {"content-digest": "sha-256=:MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=:"},
        {
            "content-digest": "sha-256=:MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=:;mooooo"
        },
        {
            "content-digest": "sha-256=:MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=:;sound=mooo;mooo"
        },
        {
            "content-digest": "sha-256=:MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=:,sha-new=:xx==:"
        },
        {
            "content-digest": "sha-512=:S3whoWr+QjPjztVYoyMrjTHT69+7rL7lKcuL72oBsgxRE396sAM+WNtjrE8cWloUUTlkKSoSDaGCaNGqBBj3qQ==:"
        },
        {
            "content-digest": "sha-256=:MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=:,sha-new=:xx==:,sha-512=:S3whoWr+QjPjztVYoyMrjTHT69+7rL7lKcuL72oBsgxRE396sAM+WNtjrE8cWloUUTlkKSoSDaGCaNGqBBj3qQ==:"
        },
    ],
)
def test_validate_digest_success(headers):
    checker = SignatureChecker(key_retriever=AsyncMock())

    assert checker.validate_digest(headers, b'{"cows": "good"}')


@pytest.mark.parametrize(
    "headers",
    [
        {"digest": "sha-256=xxxxx"},
        {"content-digest": "sha-new=:xx==:"},
        {"content-digest": "sha-256=:xx==:"},
        {"content-digest": "sha-256=:xx=:"},
        {"content-digest": "invalid"},
        {
            "content-digest": "sha-256=:MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k=:,sha-new=:xx==:,sha-512=:xx==:"
        },
    ],
)
def test_validate_digest_failure(headers):
    checker = SignatureChecker(key_retriever=AsyncMock())

    assert not checker.validate_digest(headers, b'{"cows": "good"}')



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from datetime import datetime, timedelta, timezone

from . import check_max_offset_now, get_gmt_now, parse_gmt


def test_get_gmt_now() -> None:
    date_string = get_gmt_now()

    assert "GMT" in date_string


def test_parse_gmt() -> None:
    date_string = get_gmt_now()

    parsed = parse_gmt(date_string)
    now = datetime.now(tz=timezone.utc)

    assert parsed <= now
    assert parsed >= now - timedelta(seconds=5)


def test_check_max_offset_now() -> None:
    now = datetime.now(tz=timezone.utc)

    assert check_max_offset_now(now - timedelta(minutes=4))
    assert check_max_offset_now(now + timedelta(minutes=4))
    assert not check_max_offset_now(now - timedelta(minutes=6))
    assert not check_max_offset_now(now + timedelta(minutes=6))



# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import pytest

from bovine.crypto.test import public_key, public_key_multibase
from bovine.crypto.types import CryptographicIdentifier

from . import Actor

base_actor = {
    "@context": "about:bovine",
    "id": "https://local/actor",
    "inbox": "https://local/actor/inbox",
    "outbox": "https://local/actor/outbox",
    "preferredUsername": "alice",
}


@pytest.mark.parametrize(
    ["key_data", "key_ids"],
    [
        (
            {
                "publicKey": {
                    "id": base_actor["id"] + "#key",
                    "publicKeyPem": public_key,
                    "owner": base_actor["id"],
                }
            },
            {base_actor["id"] + "#key"},
        ),
        (
            {
                "publicKey": [
                    {
                        "id": base_actor["id"] + "#key",
                        "publicKeyPem": public_key,
                        "owner": base_actor["id"],
                    }
                ]
            },
            {base_actor["id"] + "#key"},
        ),
        (
            {
                "publicKey": [
                    {
                        "id": base_actor["id"] + "#one",
                        "publicKeyPem": public_key,
                        "owner": base_actor["id"],
                    },
                    {
                        "id": base_actor["id"] + "#two",
                        "publicKeyPem": public_key,
                        "owner": base_actor["id"],
                    },
                ]
            },
            {base_actor["id"] + "#one", base_actor["id"] + "#two"},
        ),
        (
            {
                "assertionMethod": [
                    {
                        "id": base_actor["id"] + "#one",
                        "type": "Multikey",
                        "controller": base_actor["id"],
                        "publicKeyMultibase": public_key_multibase,
                    }
                ]
            },
            {base_actor["id"] + "#one"},
        ),
        (
            {
                "assertionMethod": {
                    "id": base_actor["id"] + "#one",
                    "type": "Multikey",
                    "controller": base_actor["id"],
                    "publicKeyMultibase": public_key_multibase,
                }
            },
            {base_actor["id"] + "#one"},
        ),
        (
            {
                "authentication": {
                    "id": base_actor["id"] + "#one",
                    "type": "Multikey",
                    "controller": base_actor["id"],
                    "publicKeyMultibase": public_key_multibase,
                }
            },
            {base_actor["id"] + "#one"},
        ),
        (
            {
                "verificationMethod": {
                    "id": base_actor["id"] + "#one",
                    "type": "Multikey",
                    "controller": base_actor["id"],
                    "publicKeyMultibase": public_key_multibase,
                }
            },
            {base_actor["id"] + "#one"},
        ),
    ],
)
def test_valid_cryptographic_identifier(key_data, key_ids):
    actor = Actor({**base_actor, **key_data})

    identifiers = actor.cryptographic_identifiers

    assert len(identifiers) > 0

    assert all(isinstance(x, CryptographicIdentifier) for x in identifiers)
    assert all(x.controller == base_actor["id"] for x in identifiers)

    assert {x.id for x in identifiers} == key_ids


@pytest.mark.parametrize(
    ["key_data", "relationships"],
    [
        (
            {
                "assertionMethod": {
                    "id": base_actor["id"] + "#one",
                    "type": "Multikey",
                    "controller": base_actor["id"],
                    "publicKeyMultibase": public_key_multibase,
                }
            },
            {"assertionMethod"},
        ),
        (
            {
                "authentication": {
                    "id": base_actor["id"] + "#one",
                    "type": "Multikey",
                    "controller": base_actor["id"],
                    "publicKeyMultibase": public_key_multibase,
                }
            },
            {"authentication"},
        ),
        (
            {
                "verificationMethod": {
                    "id": base_actor["id"] + "#one",
                    "type": "Multikey",
                    "controller": base_actor["id"],
                    "publicKeyMultibase": public_key_multibase,
                }
            },
            set(),
        ),
    ],
)
def test_verification_relationships(key_data, relationships):
    actor = Actor({**base_actor, **key_data})

    identifiers = actor.cryptographic_identifiers

    assert len(identifiers) == 1
    assert identifiers[0].verification_relationships == relationships



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from .link import Link


def test_object_link_fep_e232():
    link = Link.for_object("https://server.example/objects/123")

    assert link.build() == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "type": "Link",
        "mediaType": 'application/ld+json; profile="https://www.w3.org/ns/activitystreams"',
        "href": "https://server.example/objects/123",
        "name": "RE: https://server.example/objects/123",
    }



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from . import OrderedCollection


def test_ordered_collection_builder() -> None:
    result = OrderedCollection(id="url").build()

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url",
        "totalItems": 0,
        "type": "OrderedCollection",
    }


def test_ordered_collection_builder_with_items() -> None:
    result = OrderedCollection(id="url", count=1, items=[{"item": "1"}]).build()

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url",
        "orderedItems": [{"item": "1"}],
        "totalItems": 1,
        "type": "OrderedCollection",
    }


def test_ordered_collection_builder_with_fist_last() -> None:
    result = OrderedCollection(id="url", count=1, first="first", last="last").build()

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url",
        "first": "first",
        "last": "last",
        "totalItems": 1,
        "type": "OrderedCollection",
    }



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import aiohttp

from bovine import BovineClient
from bovine.clients import lookup_account_with_webfinger
from bovine.utils import now_isoformat

from .validation import validate_tos


def validate(to, public):
    if not validate_tos(to):
        print("Invalid recipients")
        exit(1)

    if len(to) == 0 and not public:
        print("Specify at least one recipient")
        exit(1)


async def resolve(session: aiohttp.ClientSession, to: str) -> str | None:
    if "@" in to:
        return await lookup_account_with_webfinger(session, to)

    return to


async def process(secret, host, to, public, message):
    if host and secret:
        client = BovineClient(host=host, secret=secret)
    else:
        client = BovineClient.from_file("bovine_user.toml")

    async with client:
        recipients = [await resolve(client.session, to) for to in to]

        mentions = [
            await client.object_factory.mention_for_actor_uri(recipient)
            for recipient in recipients
        ]
        mentions = [m.build() for m in mentions]

        note = client.object_factory.note(
            to=set(recipients),
            tag=mentions,
            content=message,
            published=now_isoformat(),
        )

        if public:
            note = note.as_public()

        create = client.activity_factory.create(note.build()).build()

        await client.send_to_outbox(create)



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

from . import JrdData, JrdLink, pydantic_to_json


def test_jrd_link():
    assert pydantic_to_json(JrdLink()) == {}
    assert pydantic_to_json(JrdLink(rel="rel", href="href", type="type")) == {
        x: x for x in ["rel", "href", "type"]
    }


def test_jrd_asdict():
    assert pydantic_to_json(JrdData()) == {}
    assert pydantic_to_json(JrdData(subject="acct:test")) == {"subject": "acct:test"}

    link = JrdLink(href="https://some.example/")

    assert pydantic_to_json(JrdData(links=[link])) == {
        "links": [{"href": "https://some.example/"}]
    }



# SPDX-FileCopyrightText: 2023 Helge
# SPDX-FileCopyrightText: 2024 helge
#
# SPDX-License-Identifier: MIT

import aiohttp
from typing import Callable, Tuple
import bovine.clients.signed_http_methods

from bovine.crypto.types import CryptographicSecret
from bovine.crypto.helper import content_digest_sha256


def legacy_digest_method(body):
    return "digest", content_digest_sha256(body)


class SignedHttpClient:
    """Client for using HTTP Signatures"""

    def __init__(
        self,
        session: aiohttp.ClientSession,
        public_key_url: str,
        private_key: str,
        digest_method: Callable[[bytes], Tuple[str, str]] | None = None,
    ):
        """Creates the http client. By using [content_digest_sha256_rfc_9530][bovine.crypto.helper.content_digest_sha256_rfc_9530] one can send the Content Digest according to RFC 9530.

        :param session: The aiohttp.ClientSession
        :param public_key_url: Used as keyId when signing
        :param private_key: The pem encoded private key
        :param digest_method: Allow specifying the method used to compute the digest
        """
        self.session = session
        self.secret = CryptographicSecret.from_pem(public_key_url, private_key)

        if digest_method is None:
            digest_method = legacy_digest_method
        self.digest_method = digest_method

    async def get(self, url: str, headers: dict = {}) -> aiohttp.ClientResponse:
        """Retrieves url using a signed get request

        :param url: URL to get
        :param headers: extra headers"""
        return await bovine.clients.signed_http_methods.signed_get(
            self.session, self.secret, url, headers
        )

    async def post(
        self, url: str, body: str, headers: dict = {}, content_type: str = None
    ):
        """Posts to url using a signed post request

        :param url: URL to post to
        :param body: The post body
        :param headers: extra headers
        :param content_type: Content type of the message"""
        return await bovine.clients.signed_http_methods.signed_post(
            self.session,
            self.secret,
            url,
            body,
            headers=headers,
            digest_method=self.digest_method,
            content_type=content_type,
        )

    def event_source(self, url: str):
        """Opens an event source to url

        :param url: Url to query"""
        return bovine.clients.signed_http_methods.signed_event_source(
            self.session, self.secret, url
        )
